from hedra.core.engines.types.common.base_result import BaseResult
from hedra.core.engines.types.common.types import RequestTypes
from .task import Task


class TaskResult(BaseResult):

    __slots__ = (
        'action_id',
        'wait_start',
        'start',
        'complete',
        'data',
        'error'
    )

    def __init__(self, task: Task, error: Exception=None):
        super(
            TaskResult,
            self
        ).__init__(
            task.action_id,
            task.name,
            task.source,
            task.metadata.user,
            task.metadata.tags,
            RequestTypes.TASK,
            task.hooks.checks,
            error
        )

        self.wait_start = 0
        self.start = 0
        self.complete = 0

        self.data = None
        self.error = error


import asyncio
from .bootstrap_services import BootstrapManager


class BootstrapServer:

    def __init__(self, config) -> None:
        self.manager = BootstrapManager(config)
        self.loop = None

    def start(self):
        self.loop = asyncio.get_event_loop()
        self.loop.run_until_complete(self.manager.start_server())

    def run(self):
        self.loop.run_until_complete(self.manager.wait())

    def kill(self):
        self.loop.run_until_complete(self.manager.stop())


import asyncio
from alive_progress import alive_bar
from hedra.core import Executor


class LocalWorker:

    def __init__(self, config):
        self.config = config
        self.reporter_config = config.reporter_config
        self.executor = None
        self.loop = asyncio.get_event_loop()

    def run(self, config=None):
        if config is None:
            config = self.config
        
        self.executor = Executor(config)
        self.loop.run_until_complete(self.executor.setup(self.reporter_config))
        if self.config.executor_config.get('persona_type') == 'timed' and self.config.runner_mode == 'local':
            with alive_bar() as bar:
                self.loop.run_until_complete(self.executor.generate_load())
                bar()

        else:
            self.loop.run_until_complete(self.executor.generate_load())

    def parse_results(self):
        return self.loop.run_until_complete(self.executor.serialize_results())

    def complete(self):
        aggregate_events = self.loop.run_until_complete(self.executor.calculate_results())
        return self.loop.run_until_complete(self.executor.submit_results(aggregate_events))
        


import importlib
import ntpath
import sys
import inspect
from collections import defaultdict
from pathlib import Path
from typing import Any, Dict
from hedra.plugins.types.plugin_types import PluginType
from hedra.plugins.types.common.plugin import Plugin
from .stage import Stage


def import_from_path(path: str) -> Any:
    package_dir = Path(path).resolve().parent
    package_dir_path = str(package_dir)
    package_dir_module = package_dir_path.split('/')[-1]
    
    package = ntpath.basename(path)
    package_slug = package.split('.')[0]
    spec = importlib.util.spec_from_file_location(f'{package_dir_module}.{package_slug}', path)
    
    if path not in sys.path:
        sys.path.append(str(package_dir.parent))

    module = importlib.util.module_from_spec(spec)

    sys.modules[module.__name__] = module

    spec.loader.exec_module(module)

    return module


def import_stages(path: str) -> Dict[str, Stage]:
    
    module = import_from_path(path)
    direct_decendants = list({cls.__name__: cls for cls in Stage.__subclasses__()}.values())

    discovered = {}
    for name, stage_candidate in inspect.getmembers(module):
        if inspect.isclass(stage_candidate) and issubclass(stage_candidate, Stage) and stage_candidate not in direct_decendants:
            discovered[name] = stage_candidate

    return discovered


def import_plugins(path: str) -> Dict[PluginType, Dict[str, Plugin]]:
    module = import_from_path(path)
    direct_decendants = list({cls.__name__: cls for cls in Plugin.__subclasses__()}.values())

    plugins_by_type = defaultdict(dict)
    for name, plugin_candidate in inspect.getmembers(module):
        if inspect.isclass(plugin_candidate) and issubclass(plugin_candidate, Plugin) and plugin_candidate not in direct_decendants:
            plugins_by_type[plugin_candidate.type][plugin_candidate.name] = plugin_candidate

    return plugins_by_type


from typing import Coroutine, Dict, Any, Callable, Awaitable, Type
from hedra.core.graphs.hooks.hook_types.hook_type import HookType
from .hook import Hook


class EventHook(Hook):

    def __init__(
        self, 
        name: str, 
        shortname: str, 
        call: Callable[..., Awaitable[Any]], 
        *names: str,
        pre: bool=False
    ) -> None:
        super().__init__(
            name, 
            shortname, 
            call, 
            hook_type=HookType.EVENT
        )

        self.call: Type[self._call] = self._call
        self.names = list(set(names))
        self.pre = pre
        self.events: Dict[str, Coroutine] = {}


"""Synchronization primitives."""

__all__ = ('Lock', 'Event', 'Condition', 'Semaphore',
           'BoundedSemaphore', 'Barrier')

import collections
from asyncio import exceptions
from asyncio import mixins

class _ContextManagerMixin:
    async def __aenter__(self):
        # We have no use for the "as ..."  clause in the with
        # statement for locks.
        return None

    async def __aexit__(self, exc_type, exc, tb):
        pass

class NoOpSemaphore(_ContextManagerMixin, mixins._LoopBoundMixin):
    """A Semaphore implementation.
    A semaphore manages an internal counter which is decremented by each
    acquire() call and incremented by each release() call. The counter
    can never go below zero; when acquire() finds that it is zero, it blocks,
    waiting until some other thread calls release().
    Semaphores also support the context management protocol.
    The optional argument gives the initial value for the internal
    counter; it defaults to 1. If the value given is less than 0,
    ValueError is raised.
    """

    def __init__(self):
        pass


import math
from .variance import (
    Variance
)

class StandardDeviation:

    __slots__ = (
        'variance'
    )

    def __init__(self):
        self.variance = Variance()

    def update(self, new_value):
        self.variance.update(new_value)
        return self

    def get(self):
        return math.sqrt(self.variance.get())



from typing import Coroutine, Dict, List
from hedra.core.engines.types.common.base_action import BaseAction
from hedra.core.engines.types.common.hooks import Hooks
from hedra.core.engines.types.common.types import RequestTypes


class Task(BaseAction):

    __slots__ = (
        'action_id',
        'protocols', 
        'name', 
        'is_setup', 
        'metadata', 
        'hooks',
        'type',
        'source',
        'execute'
    )

    def __init__(
        self, 
        name: str,
        task_action: Coroutine,
        source: str=None,
        user: str=None, 
        tags: List[Dict[str, str]] = []
    ):
        super(Task, self).__init__(
            name,
            user,
            tags
        )

        self.type = RequestTypes.TASK
        self.source = source
        self.execute = task_action
        self.hooks: Hooks[Task] = Hooks()

    def to_serializable(self):

        return {
            'name': self.name,
            'type': self.type,
            'metadata': {
                'user': self.metadata.user,
                'tags': self.metadata.tags
            },
            'hooks': self.hooks.to_serializable()
        }


from collections import deque
from typing import Dict, List, Tuple, Union
from hedra.core.engines.types.common.decoder import Decoder
from hedra.core.engines.types.common.encoder import Encoder
from .base_event import BaseEvent


def is_informational_response(headers: Tuple[bytes, bytes]):
    """
    Searches a header block for a :status header to confirm that a given
    collection of headers are an informational response. Assumes the header
    block is well formed: that is, that the HTTP/2 special headers are first
    in the block, and so that it can stop looking when it finds the first
    header field whose name does not begin with a colon.

    :param headers: The HTTP/2 header block.
    :returns: A boolean indicating if this is an informational response.
    """
    for n, v in headers:
        sigil = b':'
        status = b':status'
        informational_start = b'1'

        # If we find a non-special header, we're done here: stop looping.
        if not n.startswith(sigil):
            return False

        # This isn't the status header, bail.
        if n != status:
            continue

        # If the first digit is a 1, we've got informational headers.
        return v.startswith(informational_start)


# Parsing headers mid load-test is *expensive* so we want to defer
# this work until later.
class DeferredHeaders(BaseEvent):
    event_type='DEFERRED_HEADERS'

    __slots__ = (
        'stream_id',
        'hpack_table',
        'raw_headers',
        'stream_ended',
        'end_stream',
        'priority',
        'encoding',
        'priority_updated'
    )

    def __init__(self, encoder: Encoder, frame, encoding: Union[str, None]) -> None:
        super().__init__()
        self.stream_id = frame.stream_id
        self.hpack_table = encoder.header_table
        self.raw_headers = frame.data
        self.stream_ended = None
        self.end_stream = 'END_STREAM' in frame.flags
        self.priority = 'PRIORITY' in frame.flags
        self.encoding = encoding
        self.priority_updated = None

    def parse(self) -> Tuple[int, Dict[str, str]]:

        decoder = Decoder()
        decoder.header_table = self.hpack_table
        decoder.header_table_size = self.hpack_table.maxsize
        headers: List[Tuple[bytes, bytes]] = decoder.decode(self.raw_headers, raw=True)

        header_encoding = self.encoding
        if header_encoding:
            decoded_headers = []
            
            for header in headers:
                name, value = header
                decoded_headers.append(header.__class__(
                    name.decode(header_encoding),
                    value.decode(header_encoding)
                ))

            headers = decoded_headers

        status_code = None
        headers_dict = {}
        for k, v in headers:
            if k == b":status":
                status_code = int(v.decode("ascii", errors="ignore"))
            elif k.startswith(b":"):
                headers_dict[k.strip(b':')] = v
            else:
                headers_dict[k] = v
        
        return status_code, headers_dict

    


import time

from hedra.core.personas.parallel.time import parse_time

import time
import asyncio


class Job:

    def __init__(self, job_id, pipeline, config) -> None:

        self.config = config
        self.job_id = job_id
        self.start = None
        self.elapsed = 0
        self.timeout = pipeline.timeout
        
        self.task = None
        self.results_task = None
        self.registered_workers = []
        self.excluded_workers = []
        self.pipeline = pipeline
        self.results = None

        self.status = None

    async def set_job_status(self, status):
        self.status = status

    async def start_timer(self):
        self.start = time.time()

    async def update_time_elapsed(self):
        self.elapsed = time.time() - self.start

    async def check_if_deadline_exceeded(self):
        return self.elapsed > self.timeout

    async def attach_worker_to_job(self, leader_address):
        self.registered_workers += [leader_address]
        self.registered_workers = list(set(self.registered_workers))

    async def attach_worker_to_excluded(self, leader_address):
        self.excluded_workers += [leader_address]
        self.excluded_workers = list(set(self.excluded_workers))

    async def remove_leader(self, leader_address):
        if leader_address in self.registered_workers:
            self.registered_workers.remove(leader_address)

    async def get_task(self):
        if self.results is None:
            self.results = await self.task
        
        return self.results






import uuid
from typing import Dict, List, Any
from hedra.core.engines.types.common.metadata import Metadata
from .hooks import Hooks


class Page:

    __slots__ = (
        'selector',
        'attribute',
        'x_coordinate',
        'y_coordinate',
        'frame'
    )

    def __init__(self, selector: str=None, attribute: str=None, x_coordinate: int=0, y_coordinate: int=0, frame=0) -> None:
        self.selector = selector
        self.attribute = attribute
        self.x_coordinate = x_coordinate
        self.y_coordinate = y_coordinate
        self.frame = frame

    def to_serializable(self):
        return {
            'selector': self.selector,
            'attribute': self.attribute,
            'x_coordinage': self.x_coordinate,
            'y_coordinate': self.y_coordinate,
            'frame': self.frame
        }


class URL:

    __slots__ = (
        'location',
        'headers'
    )

    def __init__(self, location: str=None, headers: Dict[str, str]={}) -> None:
        self.location = location
        self.headers = headers

    def to_serializable(self):
        return {
                'location': self.location,
                'headers': self.headers
        }


class Input:

    __slots__ = (
        'key',
        'text',
        'expression',
        'args',
        'filepath',
        'file',
        'path',
        'option',
        'by_label',
        'by_value'
    )

    def __init__(
        self, 
        key=None, 
        text=None, 
        expression: str=None, 
        args: List[Any]=None, 
        filepath=None, 
        file=None, 
        path: str=None,
        option: Any=None,
        by_label: bool=False,
        by_value: bool=False
    ) -> None:
        self.key = key
        self.text = text
        self.expression = expression
        self.args = args
        self.filepath = filepath
        self.file = file
        self.path = path
        self.option = option
        self.by_label = by_label
        self.by_value = by_value

    def to_serializable(self):
        return {
                'key': self.key,
                'text': self.text,
                'expression': self.expression,
                'args': self.args,
                'filepath': self.filepath,
                'file': self.file,
                'path': self.path,
                'option': self.option,
                'by_label': self.by_label,
                'by_value': self.by_value
        }


class Options:

    __slots__ = (
        'event',
        'option',
        'is_checked',
        'timeout',
        'extra',
        'switch_by'
    )

    def __init__(self, event: str=None, option=None, is_checked=False, timeout=10, extra={}, switch_by: str='url') -> None:
        self.event = event
        self.option = option
        self.is_checked = is_checked
        self.timeout = timeout
        self.extra = extra
        self.switch_by = switch_by

    def to_serializable(self):
        return {
            'event': self.event,
            'option': self.option,
            'is_checked': self.is_checked,
            'timeout': self.timeout,
            'extra': self.extra,
            'switch_by': self.switch_by
        }

class PlaywrightCommand:

    __slots__ = (
        'action_id'
        'name',
        'command',
        'page',
        'url',
        'input',
        'options',
        'metadata',
        'hooks'
    )
    
    def __init__(self, 
        name, 
        command, 
        page: Page = Page(), 
        url: URL = URL(), 
        input: Input = Input(), 
        options: Options = Options(), 
        user: str = None, 
        tags: List[Dict[str, str]] = []
    ) -> None:
        self.action_id = str(uuid.uuid4())
        self.name = name
        self.command = command
        self.page = page
        self.url = url
        self.input = input
        self.options = options
        self.metadata = Metadata(user, tags)
        self.hooks: Hooks[PlaywrightCommand] = Hooks()

    def to_serializable(self):

        return {
            'name': self.name,
            'command': self.command,
            'page': self.page.to_serializable(),
            'url': self.url.to_serializable(),
            'input': self.input.to_serializable(),
            'options': self.options.to_serializable(),
            'metadata': {
                'user': self.metadata.user,
                'tags': self.metadata.tags
            },
            'hooks': self.hooks.to_serializable()
        }

    def iter_values(self):
        return {
            'name': self.name,
            'command': self.command,
            **self.page.to_serializable(),
            **self.url.to_serializable()
            **self.input.to_serializable(),
            **self.options.to_serializable()
        }


import math
import asyncio
import psutil
from hedra.core.personas.types.default_persona.default_persona import DefaultPersona
from hedra.core.personas.batching.param_type import ParamType
from hedra.core.engines.client.config import Config
from hedra.core.personas.types.default_persona import DefaultPersona
from hedra.core.personas.batching.batch import Batch
from typing import Dict, List, Tuple, Union


class BaseAlgorithm:

    def __init__(
        self, 
        config: Dict[str, Union[List[Tuple[Union[int, float]]], int, Config]]
    ) -> None:
        self.stage_config: Config = config.get('stage_config')
        self.max_iter = config.get('iterations', 10)
        self.params = config.get('params', {})
        self.persona_total_time = self.stage_config.total_time
        self.batch = Batch(self.stage_config)
        self.current_params = {}

        self.batch_time = self.stage_config.total_time/self.max_iter

        self.param_names = []
        self.bounds = []

        self.param_values: Dict[str, Union[float, int]] = self.get_params()
        self.current_params = {}

        optimize_params = [
            param_name for param_name in self.params.keys() if param_name in self.param_values
        ]

        if len(optimize_params) < 1:
            optimize_params = [
                'batch_size'
            ]

        for param_name in optimize_params:
            min_range, max_range = self.params.get(param_name)
            param = self.param_values.get(param_name, {})
            value = param.get('value')
            params_type = param.get('type')

            self.param_names.append(param_name)

            if max_range <= min_range:
                raise Exception(
                    f'Err. - maximum value of optimization parameter {param_name} must be greater than minimum value.'
                )

            if params_type == ParamType.INTEGER:
                min_range = math.floor(min_range * value)
                max_range = math.ceil(max_range * value)

            else:
                min_range = min_range * value
                max_range = max_range * value
            
            self.bounds.append((
                min_range,
                max_range
            ))

            self.current_params[param_name] = value
            
        self.fixed_iters = False
        self.iters = 0

    def get_params(self):
        return self.batch.to_params()

    def update_params(self, persona: DefaultPersona) -> DefaultPersona:

        persona.total_time = self.batch_time

        batch_size = int(self.current_params.get(
            'batch_size',
            persona.batch.size
        ))

        batch_interval = float(self.current_params.get(
            'batch_interval',
            persona.batch.interval
        ))

        batch_gradient = float(self.current_params.get(
            'batch_gradient',
            persona.batch.gradient
        ))

        for hook in persona._hooks:

            if batch_size <= psutil.cpu_count():
                batch_size = 1000

            hook.session.pool.size = batch_size
            hook.session.sem = asyncio.Semaphore(batch_size)
            hook.session.pool.connections = []
            hook.session.pool.create_pool()

        persona.batch.size = batch_size
        persona.batch.interval = batch_interval
        persona.batch.gradient = batch_gradient

        return persona

    def optimize(self, func):
        raise NotImplementedError(
            'Err. - Base Algorithm is an abstract class and its optimize method should be overridden.'
        )


import os
from typing import List
import uuid
import grpc
from broadkast.grpc import BroadkastClient
from broadkast.discovery import Query
from hedra.runners.bootstrap_services.proto.bootstrap_pb2 import ServiceRegistrationStatusResponse
from hedra.runners.leader_services.bootstrap.discovered_leaders_registry import DiscoveredLeadersRegistry
from hedra.runners.utils.connect_timeout import connect_with_retry_async
from hedra.runners.bootstrap_services.proto import (
    BootstrapServerStub,
    ServiceRegisterRequest,
    ServiceRegistrationStatusRequest,
    ServiceRegistrationStatusResponse
)
from broadkast.grpc.proto import (
    ServiceDiscoveryResponse
)


class BoostrapClient:

    def __init__(self, config) -> None:

        self.use_bootstrap = config.distributed_config.get(
            'use_bootstrap'
        )
        
        self.bootsrap_ip = config.distributed_config.get(
            'bootstrap_ip',
            os.getenv('BOOSTRAP_SERVER_IP')
        )

        self.bootstrap_port = config.distributed_config.get(
            'bootstrap_port',
            os.getenv('BOOTSTRAP_SERVER_PORT', "8711")
        )


        if self.bootsrap_ip and self.bootstrap_port:
            self.bootstrap_service_address = f'{self.bootsrap_ip}:{self.bootstrap_port}'

        else:
            self.bootstrap_service_address = config.distributed_config.get(
                'bootstrap_address',
                os.getenv(
                    'BOOTSTRAP_SERVER_ADDRESS'
                )
            )

        self.registry = DiscoveredLeadersRegistry()
        self.channel = grpc.aio.insecure_channel(self.bootstrap_service_address)
        self.service = BootstrapServerStub(self.channel)
        self.broadkast_client = BroadkastClient(config=config.distributed_config)

    @connect_with_retry_async(wait_buffer=5, timeout_threshold=15)
    async def register_self(self, leader_address):
        registration_request: ServiceRegistrationStatusResponse = ServiceRegisterRequest(
            service_address=leader_address,
            service_name='hedra_leader'
        )

        registration_response = await self.service.RegisterService(registration_request)

        await self.registry.register(registration_response, host=True)

    @connect_with_retry_async(wait_buffer=5, timeout_threshold=15)
    async def get_registration_status(self, leader_address, leader_id):
        registration_status_request = ServiceRegistrationStatusRequest(
            service_address=leader_address,
            service_id=str(leader_id),
            service_name='hedra_leader'
        )

        status_response = await self.service.GetServiceRegistrationStatus(registration_status_request)

        await self.registry.register(status_response, host=False)

    @connect_with_retry_async(wait_buffer=5, timeout_threshold=15)
    async def discover(self, host_address: str=None) -> None:
        discovered: ServiceDiscoveryResponse = await self.broadkast_client.discover_service(
            'hedra_leader',
            query=Query(
                labels=[
                    {
                        'name': 'app',
                        'value': 'hedra'
                    },
                    {
                        'name': 'role',
                        'value': 'leader'
                    }
                ]
            ),
            service_port='6669'
        )
        
        await self.registry.register_kubernetes(discovered, host=host_address)

    




import asyncio
import time
import uuid
from hedra.core.engines.types.http2.client import MercuryHTTP2Client
from hedra.core.engines.types.common.timeouts import Timeouts
from typing import Awaitable, Union
from .action import GraphQLHTTP2Action
from .result import GraphQLHTTP2Result

GraphQLHTTP2ResponseFuture = Awaitable[Union[GraphQLHTTP2Action, Exception]]


class MercuryGraphQLHTTP2Client(MercuryHTTP2Client):

    def __init__(
        self, 
        concurrency: int = 10 ** 3, 
        timeouts: Timeouts = Timeouts(), 
        reset_connections: bool = False
    ) -> None:

        super(
            MercuryGraphQLHTTP2Client, 
            self
        ).__init__(
            concurrency, 
            timeouts, 
            reset_connections
        )

        self.session_id = str(uuid.uuid4())

    async def execute_prepared_request(self, action: GraphQLHTTP2Action) -> GraphQLHTTP2Action:
        
        response = GraphQLHTTP2Result(action)
        response.wait_start = time.monotonic()
        self.active += 1

        async with self.sem:
            stream = self.pool.streams.pop()
            connection = self.pool.connections.pop()
        
            try:

                if action.hooks.listen:
                    event = asyncio.Event()
                    action.hooks.channel_events.append(event)
                    await event.wait()

                
                if action.hooks.before:
                    action = await action.hooks.before(action, response)
                    action.setup()

                response.start = time.monotonic()

                reader_writer = await asyncio.wait_for(stream.connect(
                    action.url.hostname,
                    action.url.ip_addr,
                    action.url.port,
                    action.url.socket_config,
                    ssl=action.ssl_context
                ), self.timeouts.connect_timeout)

                reader_writer.encoder = action.hpack_encoder
                connection.connect(reader_writer)

                response.connect_end = time.monotonic()

                connection.send_request_headers(action, reader_writer)

                if action.encoded_data is not None:
                    await connection.submit_request_body(action, reader_writer)

                response.write_end = time.monotonic()

                await connection.receive_response(response, reader_writer)

                response.complete = time.monotonic()

                if action.hooks.after:
                    action = await action.hooks.after(action, response)
                    action.setup()

                if action.hooks.notify:
                    await asyncio.gather(*[
                        asyncio.create_task(
                            channel(response, action.hooks.listeners)
                        ) for channel in action.hooks.channels
                    ])

                    for listener in action.hooks.listeners: 
                        if len(listener.hooks.channel_events) > 0:
                            listener.setup()
                            event = listener.hooks.channel_events.pop()
                            if not event.is_set():
                                event.set()       

            
                self.pool.streams.append(stream)
                self.pool.connections.append(connection)
                
            except Exception as e:
                response.response_code = 500
                response.error = str(e)

                self.pool.reset()

            self.active -= 1
            if self.waiter and self.active <= self.pool.size:

                try:
                    self.waiter.set_result(None)
                    self.waiter = None

                except asyncio.InvalidStateError:
                    self.waiter = None

            return response


from typing import List


class InvalidPluginType(Exception):

    def __init__(self, plugin_type: str, valid_types: List[str]) -> None:

        valid_plugin_types = '\n-'.join(valid_types)

        super().__init__(
            f'\n\nError - invalid plugin type - {plugin_type} - specified.\n\nValid types are: \n-{valid_plugin_types}\n'
        )


from typing import Dict
from pydantic import BaseModel
from hedra.reporting.types.common.types import ReporterTypes


class TelegrafStatsDConfig(BaseModel):
    host: str='0.0.0.0'
    port: int=8125
    custom_fields: Dict[str, str]={}
    reporter_type: ReporterTypes=ReporterTypes.TelegrafStatsD


from hedra.core.graphs.hooks.hook_types.internal import Internal
from hedra.core.graphs.stages.types.stage_types import StageTypes
from hedra.core.graphs.stages.base.stage import Stage


class Wait(Stage):
    stage_type=StageTypes.WAIT

    def __init__(self) -> None:
        super().__init__()

    @Internal()
    async def run(self):
        await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Waiting for Stage group to complete')
        await self.logger.spinner.append_message(f'Stage - {self.name} - Waiting for Stage group to complete')
        



from hedra.core.engines.types.task.result import TaskResult
from .base_event import BaseEvent


class TaskEvent(BaseEvent):

    def __init__(self, result: TaskResult) -> None:
        super().__init__(result)

        self.time = result.complete - result.start
        self.timings = {
            'total': self.time,
            'waiting': result.start - result.wait_start,
            'reading': result.complete - result.write_end
        }

        self.data = result.data


import click


@click.group(help="Commands to run graphs on and manage distributed instances of Hedra.")
def cloud():
    pass


@cloud.group(help='Login to a Hedra Cloud cluster.')
@click.option(
    '--username',
    help='Hedra Cloud cluster username.'
)
@click.option(
    '--token',
    help='Hedra Cloud cluster auth token.'
)
@click.option(
    '--cluster',
    help='Hedra Cluster address.'
)
def login(username: str, auth_token: str, cluster: str):
    pass


@cloud.group(help='Run the specified graph on the Hedra Cloud cluster.')
@click.argument('graph_name')
@click.option(
    '--cluster',
    help='Hedra Cluster address.'
)
def run(graph_name: str, cluster: str):
    pass


@cloud.group(help='Trigger a project sync on the Hedra Cloud cluster.')
@click.option(
    '--cluster',
    help='Hedra Cluster address.'
)
def sync(cluster: str):
    pass


@cloud.group(help='Submit a test config to the Hedra Cloud cluster.')
@click.argument('config_path')
@click.option(
    '--cluster',
    help='Hedra Cluster address.'
)
def submit(config_path: str, cluster: str):
    pass


@cloud.group(help='Validate the specified graph on the Hedra Cloud cluster.')
@click.argument('graph_name')
@click.option(
    '--cluster',
    help='Hedra Cluster address.'
)
def check(graph_name: str, cluster: str):
    pass


@cloud.group(help='Follow test progress as it runs on the specified Hedra Cloud cluster.')
@click.argument('test_id')
@click.option(
    '--cluster',
    help='Hedra Cluster address.'
)
def watch(test_id: str, cluster: str):
    pass


@cloud.group(help='Update graph config for the specified graph on the specified Hedra Cloud cluster.')
@click.argument('graph_name')
@click.option(
    '--cluster',
    help='Hedra Cluster address.'
)
def update(graph_name: str, cluster: str):
    pass


import uuid
from typing import List
from numpy import float32, float64, int16, int32, int64
from hedra.logging import HedraLogger
from hedra.reporting.events.types.base_event import BaseEvent
from hedra.reporting.metric import MetricsSet


try:
    import sqlalchemy
    from sqlalchemy.schema import CreateTable
    from sqlalchemy.ext.asyncio import create_async_engine
    from .sqlite_config import SQLiteConfig
    has_connector = True

except Exception:
    ASYNCIO_STRATEGY = None
    sqlalchemy = None
    SQLiteConfig = None
    CreateTable = None
    OperationalError = None
    has_connector = False



class SQLite:

    def __init__(self, config: SQLiteConfig) -> None:
        self.path = f'sqlite+aiosqlite:///{config.path}'
        self.events_table_name = config.events_table
        self.metrics_table_name = config.metrics_table
        self.shared_metrics_table_name = 'stage_metrics'
        self.errors_table_name = 'stage_errors'
        self.custom_fields = config.custom_fields
        self.metadata = sqlalchemy.MetaData()

        self.database = None
        self._engine = None
        self._connection = None

        self._events_table = None
        self._metrics_table = None
        self._shared_metrics_table = None
        self._custom_metrics_tables = {}
        self._errors_table = None

        self.session_uuid = str(uuid.uuid4())
        self.metadata_string: str = None
        self.logger = HedraLogger()
        self.logger.initialize()


    async def connect(self):
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Connecting to SQLite at - {self.path} - Database: {self.database}')
        self._engine = create_async_engine(self.path)

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Connected to SQLite at - {self.path} - Database: {self.database}')
    
    async def submit_events(self, events: List[BaseEvent]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Events to Table - {self.events_table_name}')

        async with self._engine.begin() as connection:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Events to Table - {self.events_table_name} - Initiating transaction')

            for event in events:

                if self._events_table is None:
                    await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Events table - {self.events_table_name} - if not exists')

                    events_table = sqlalchemy.Table(
                        self.events_table_name,
                        self.metadata,
                        sqlalchemy.Column('id', sqlalchemy.INTEGER, primary_key=True,),
                        sqlalchemy.Column('name', sqlalchemy.TEXT),
                        sqlalchemy.Column('stage', sqlalchemy.TEXT),
                        sqlalchemy.Column('time', sqlalchemy.REAL),
                        sqlalchemy.Column('succeeded', sqlalchemy.INTEGER)
                    )

                    
                    await connection.execute(CreateTable(events_table, if_not_exists=True))
                    self._events_table = events_table

                    await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Created or set Events table - {self.events_table_name}')
            
                await connection.execute(self._events_table.insert(values=event.record))
            
            await connection.commit()
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Events to Table - {self.events_table_name} - Transaction committed')

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Events to Table - {self.events_table_name}')

    async def submit_common(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Shared Metrics to Table - {self.shared_metrics_table_name}')

        async with self._engine.begin() as connection:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Shared Metrics to Table - {self.shared_metrics_table_name} - Initiating transaction')

            if self._shared_metrics_table is None:
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Shared Metrics table - {self.shared_metrics_table_name} - if not exists')

                stage_metrics_table = sqlalchemy.Table(
                    self.shared_metrics_table_name,
                    self.metadata,
                    sqlalchemy.Column('id', sqlalchemy.INTEGER, primary_key=True),
                    sqlalchemy.Column('name', sqlalchemy.TEXT),
                    sqlalchemy.Column('stage', sqlalchemy.TEXT),
                    sqlalchemy.Column('group', sqlalchemy.TEXT),
                    sqlalchemy.Column('total', sqlalchemy.INTEGER),
                    sqlalchemy.Column('succeeded', sqlalchemy.INTEGER),
                    sqlalchemy.Column('failed', sqlalchemy.INTEGER),
                    sqlalchemy.Column('actions_per_second', sqlalchemy.REAL)
                )

                await connection.execute(CreateTable(stage_metrics_table, if_not_exists=True))
                self._shared_metrics_table = stage_metrics_table

                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Created or set Shared Metrics table - {self.shared_metrics_table_name}')

            for metrics_set in metrics_sets:
                await connection.execute(
                    self._shared_metrics_table.insert(values={
                        'name': metrics_set.name,
                        'stage': metrics_set.stage,
                        'group': 'common',
                        **metrics_set.common_stats
                    })
                )

            await connection.commit()
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Shared Metrics to Table - {self.shared_metrics_table_name} - Transaction committed')

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Shared Metrics to Table - {self.shared_metrics_table_name}')

    async def submit_metrics(self, metrics: List[MetricsSet]):
        
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Metrics to Table - {self.metrics_table_name}')

        async with self._engine.begin() as connection:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Metrics to Table - {self.metrics_table_name} - Initiating transaction')

            for metrics_set in metrics:
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Shared Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

                if self._metrics_table is None:
                    await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Metrics table - {self.metrics_table_name} - if not exists')

                    metrics_table = sqlalchemy.Table(
                        self.metrics_table_name,
                        self.metadata,
                        sqlalchemy.Column('id', sqlalchemy.INTEGER, primary_key=True),
                        sqlalchemy.Column('name', sqlalchemy.TEXT),
                        sqlalchemy.Column('stage', sqlalchemy.TEXT),
                        sqlalchemy.Column('group', sqlalchemy.TEXT),
                        sqlalchemy.Column('median', sqlalchemy.REAL),
                        sqlalchemy.Column('mean', sqlalchemy.REAL),
                        sqlalchemy.Column('variance', sqlalchemy.REAL),
                        sqlalchemy.Column('stdev', sqlalchemy.REAL),
                        sqlalchemy.Column('minimum', sqlalchemy.REAL),
                        sqlalchemy.Column('maximum', sqlalchemy.REAL)
                    )

                    for quantile in metrics_set.quantiles:
                        metrics_table.append_column(
                            sqlalchemy.Column(f'{quantile}', sqlalchemy.REAL)
                        )

                    for custom_field_name, sql_alchemy_type in metrics_set.custom_schemas:
                        metrics_table.append_column(custom_field_name, sql_alchemy_type)   
                    
                    await connection.execute(CreateTable(metrics_table, if_not_exists=True))
                    self._metrics_table = metrics_table

                    await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Created or set Metrics table - {self.metrics_table_name}')

                for group_name, group in metrics_set.groups.items():
                    await connection.execute(self._metrics_table.insert(values={
                        **group.record,
                        'group': group_name
                    }))

            await connection.commit()
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Metrics to Table - {self.metrics_table_name} - Transaction committed')

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Metrics to Table - {self.metrics_table_name}')

    async def submit_custom(self, metrics_sets: List[MetricsSet]):

        async with self._engine.begin() as connection:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Custom Metrics - Initiating transaction')
            
            for metrics_set in metrics_sets:
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Custom Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

                for custom_group_name, group in metrics_set.custom_metrics.items():
                    custom_table_name = f'{custom_group_name}_metrics'

                    if self._custom_metrics_tables.get(custom_table_name) is None:
                        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Custom Metrics table - {custom_group_name} - if not exists')

                        custom_metrics_table = sqlalchemy.Table(
                            custom_table_name,
                            self.metadata,
                            sqlalchemy.Column('id', sqlalchemy.INTEGER, primary_key=True),
                            sqlalchemy.Column('name', sqlalchemy.TEXT),
                            sqlalchemy.Column('stage', sqlalchemy.TEXT),
                            sqlalchemy.Column('group', sqlalchemy.TEXT),
                        )

                        for field, value in group.items():

                            if isinstance(value, (int, int16, int32, int64)):
                                custom_metrics_table.append_column(
                                    sqlalchemy.Column(field, sqlalchemy.INTEGER)
                                )

                            elif isinstance(value, (float, float32, float64)):
                                custom_metrics_table.append_column(
                                    sqlalchemy.Column(field, sqlalchemy.REAL)
                                )

                        await connection.execute(CreateTable(custom_metrics_table, if_not_exists=True))
                        self._custom_metrics_tables[custom_table_name] = custom_metrics_table

                        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Created or set Custom Metrics table - {custom_group_name}')

                    await connection.execute(
                        self._custom_metrics_tables[custom_table_name].insert(values={
                            'name': metrics_set.name,
                            'stage': metrics_set.stage,
                            'group': custom_group_name,
                            **group
                        })
                    )

            await connection.commit()
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Custom Metrics - Transaction committed')

    async def submit_errors(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Error Metrics to Table - {self.errors_table_name}')

        async with self._engine.begin() as connection:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Error Metrics to Table - {self.errors_table_name} - Initiating transaction')

            for metrics_set in metrics_sets:
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Error Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

                if self._errors_table is None:
                    await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Error Metrics table - {self.errors_table_name} - if not exists')

                    metrics_errors_table = sqlalchemy.Table(
                        self.errors_table_name,
                        self.metadata,
                        sqlalchemy.Column('id', sqlalchemy.INTEGER, primary_key=True),
                        sqlalchemy.Column('name', sqlalchemy.TEXT),
                        sqlalchemy.Column('stage', sqlalchemy.TEXT),
                        sqlalchemy.Column('error_message', sqlalchemy.TEXT),
                        sqlalchemy.Column('error_count', sqlalchemy.INTEGER)
                    ) 

                    await connection.execute(CreateTable(metrics_errors_table, if_not_exists=True))

                    self._errors_table = metrics_errors_table
                    await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Created or set Error Metrics table - {self.errors_table_name}')

                for error in metrics_set.errors:
                    await connection.execute(
                        self._errors_table.insert(values={
                            'name': metrics_set.name,
                            'stage': metrics_set.stage,
                            'error_message': error.get('message'),
                            'error_count': error.get('count')
                        })
                    )

            await connection.commit()
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Error Metrics to Table - {self.errors_table_name} - Transaction committed')

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Error Metrics to Table - {self.errors_table_name}')

    async def close(self):
        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Closing session - {self.session_uuid}')



import asyncio
import dill
import time
from collections import defaultdict
from typing import Dict, List, Tuple, Union
from hedra.core.graphs.events import Event
from hedra.core.graphs.hooks.hook_types.hook_type import HookType
from hedra.core.graphs.hooks.registry.registry_types import (
    EventHook, 
    ContextHook,
    ActionHook,
    TaskHook
)
from hedra.core.graphs.stages.types.stage_types import StageTypes
from hedra.core.graphs.hooks.hook_types.internal import Internal
from hedra.core.engines.client.time_parser import TimeParser
from hedra.core.graphs.stages.execute import Execute
from hedra.core.graphs.stages.base.stage import Stage
from .parallel import optimize_stage


class Optimize(Stage):
    stage_type=StageTypes.OPTIMIZE
    optimize_iterations=0
    algorithm='shg'
    stage_time_limit='1m'
    optimize_params={
        'batch_size': (0.5, 2)
    }
    
    def __init__(self) -> None:
        super().__init__()
        self.generation_optimization_candidates = 0
        self.execution_stage_id = 0

        self.results = None

        time_parser = TimeParser(self.stage_time_limit)
        self.time_limit = time_parser.time
        self.requires_shutdown = True
        self.allow_parallel = True

        self.optimization_execution_time = 0
        self.accepted_hook_types = [ HookType.EVENT, HookType.CONTEXT ]

    @Internal()
    async def run(self, stages: Dict[str, Execute]):

        events: List[Union[EventHook, Event]] = [event for event in self.hooks[HookType.EVENT]]
        pre_events: List[EventHook] = [
            event for event in events if isinstance(event, EventHook) and event.pre
            ]
        
        if len(pre_events) > 0:
            pre_event_names = ", ".join([
                event.shortname for event in pre_events
            ])

            await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Executing PRE events - {pre_event_names}')
            await asyncio.wait([
                asyncio.create_task(event.call()) for event in pre_events
            ], timeout=self.stage_timeout)

        optimization_execution_time_start = time.monotonic()

        stage_names = ', '.join(list(stages.keys()))
        stages_count = len(stages)

        await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Optimizing stages {stage_names} using {self.algorithm} algorithm')
        await self.logger.spinner.append_message(f'Optimizer - {self.name} optimizing stages {stage_names} using {self.algorithm} algorithm')

        optimization_results = []

        # We may have less workers available during the optimize stage than assigned
        # to the execute stage, so store the original workers count for later.
        stage_workers_map = {
            stage.name: stage.workers for stage in stages.values()
        }

        optimize_stages = [(
            stage.name, 
            stage
        ) for stage in stages.values()]

        batched_stages: List[Tuple[str, Execute, int]] = list(self.executor.partion_stage_batches(optimize_stages))
        batched_configs = []

        await self.logger.filesystem.aio['hedra.core'].debug(f'{self.metadata_string} - Batching optimization for - {stages_count} stages')

        for stage_name, stage, assigned_workers_count in batched_stages:

            configs = []
            stage_config = stage.client._config

            batch_size = int(stage_config.batch_size/assigned_workers_count)

            for worker_idx in range(assigned_workers_count):
                
                execute_stage_actions: List[ActionHook] = [hook.name for hook in stage.hooks[HookType.ACTION]]
                execute_stage_tasks: List[TaskHook] = [hook.name for hook in stage.hooks[HookType.TASK]]

                execute_stage_plugins = defaultdict(list)

                for plugin in stage.plugins.values():
                    execute_stage_plugins[plugin.type].append(plugin.name)

                configs.append({
                    'graph_name': self.graph_name,
                    'graph_path': self.graph_path,
                    'graph_id': self.graph_id,
                    'optimize_params': self.optimize_params,
                    'worker_idx': worker_idx,
                    'source_stage_name': self.name,
                    'source_stage_id': self.stage_id,
                    'execute_stage_name': stage_name,
                    'execute_stage_generation_count': assigned_workers_count,
                    'execute_stage_id': stage.execution_stage_id,
                    'execute_stage_config': stage_config,
                    'execute_stage_batch_size': batch_size,
                    'execute_stage_plugins': execute_stage_plugins,
                    'optimizer_iterations': self.optimize_iterations,
                    'optimizer_algorithm': self.algorithm,
                    'execute_stage_hooks': [
                        *execute_stage_actions,
                        *execute_stage_tasks
                    ],
                    'time_limit': self.time_limit
                })

            configs[assigned_workers_count-1]['execute_stage_batch_size'] += batch_size%assigned_workers_count

            configs = [
                dill.dumps(config) for config in configs
            ]

            batched_configs.append((
                stage_name,
                assigned_workers_count,
                configs
            ))

            await self.logger.filesystem.aio['hedra.core'].debug(f'{self.metadata_string} - Provisioned - {assigned_workers_count} - workers for stage - {stage_name}')

        await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Starting optimizaiton for - {stages_count} - stages')

        results = await self.executor.execute_batches(
            batched_configs,
            optimize_stage
        )

        await self.logger.filesystem.aio['hedra.core'].debug(f'{self.metadata_string} - Completed optimizaiton for - {stages_count} - stages')

        for _, result in results:
            optimization_results.extend(result)

        optimized_batch_sizes = []
        for optimization_result in optimization_results:
            optimized_config = optimization_result.get('config')
            optimized_batch_sizes.append(
                optimized_config.batch_size
            )

        optimized_batch_size = sum(optimized_batch_sizes)

        stage_optimzations = {}
        for optimization_result in optimization_results:
            
            stage_name = optimization_result.get('stage')
            optimized_config = optimization_result.get('config')

            stage = stages.get(stage_name)
            stage.client._config = optimized_config

            action_hooks: List[ActionHook] = stage.hooks[HookType.ACTION]
            for hook in action_hooks:
                hook.session.pool.size = optimized_batch_size
                hook.session.sem = asyncio.Semaphore(optimized_config.batch_size)
                hook.session.pool.connections = []
                hook.session.pool.create_pool()

            stage_optimzations[stage_name] = optimized_batch_size

            await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Stage - {stage_name} - configured to use optimized batch size of - {optimized_batch_size} - VUs')

            optimized_config.optimized = True
            stage.optimized = True

        for stage in stages.values():
            stage.workers = stage_workers_map.get(stage.name)

        self.optimization_execution_time = round(time.monotonic() - optimization_execution_time_start)

        optimized_batch_sizes = ', '.join([
            f'{stage_name}: {optimized_batch_size}' for stage_name, optimized_batch_size in stage_optimzations.items()
        ])

        post_events: List[EventHook] = [
            event for event in events if isinstance(event, EventHook) and event.pre is False
        ]

        if len(post_events) > 0:
            post_event_names = ", ".join([
                event.shortname for event in post_events
            ])

            await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Executing POST events - {post_event_names}')
            await asyncio.wait([
                asyncio.create_task(event.call()) for event in post_events
            ], timeout=self.stage_timeout)

        await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Optimization complete for stages - {stage_names} - over - {self.optimization_execution_time} - seconds')
        await self.logger.spinner.set_default_message(f'Optimized - batch sizes for stages - {optimized_batch_sizes} - over {self.optimization_execution_time} seconds')

        context_hooks: List[ContextHook] = self.hooks[HookType.CONTEXT]
        context_hooks: List[ContextHook] = self.hooks[HookType.CONTEXT]
        await asyncio.gather(*[
            asyncio.create_task(context_hook.call(self.context)) for context_hook in context_hooks
        ])
        
        return [
            result.get('params') for result in optimization_results
        ]






import asyncio
import time
import dill
import json
import threading
import os
from collections import defaultdict
from typing import Any, Dict, List
from hedra.core.engines.types.common.base_result import BaseResult
from hedra.core.graphs.hooks.registry.registrar import registrar
from hedra.logging import (
    HedraLogger,
    LoggerTypes,
    logging_manager
)
from hedra.reporting.events import EventsGroup


async def process_batch(
        stage_name: str, 
        custom_metric_hook_names: List[str],
        results_batch: List[BaseResult],
        metadata_string: str
    ):

        hedra_config_filepath = os.path.join(
            os.getcwd(),
            '.hedra.json'
        )

        hedra_config = {}
        if os.path.exists(hedra_config_filepath):
            with open(hedra_config_filepath, 'r') as hedra_config_file:
                hedra_config = json.load(hedra_config_file)

        logging_config = hedra_config.get('logging', {})
        logfiles_directory = logging_config.get(
            'logfiles_directory',
            os.getcwd()
        )

        log_level = logging_config.get('log_level', 'info')

        logging_manager.disable(
            LoggerTypes.DISTRIBUTED,
            LoggerTypes.DISTRIBUTED_FILESYSTEM
        )

        logging_manager.update_log_level(log_level)
        logging_manager.logfiles_directory = logfiles_directory


        logger = HedraLogger()
        logger.initialize()
        await logger.filesystem.aio.create_logfile('hedra.core.log')
        await logger.filesystem.aio.create_logfile('hedra.reporting.log')

        logger.filesystem.create_filelogger('hedra.core.log')
        logger.filesystem.create_filelogger('hedra.reporting.log')

        await logger.filesystem.aio['hedra.core'].info(f'{metadata_string} - Initializing results aggregation')

        start = time.monotonic()

        events =  defaultdict(EventsGroup)

        custom_metric_hooks = []
        for metric_hook_name in custom_metric_hook_names:
            custom_metric_hook = registrar.all.get(metric_hook_name)
            custom_metric_hooks.append(custom_metric_hook)

        await asyncio.gather(*[
            asyncio.create_task(events[stage_result.name].add(
                    stage_result,
                    stage_name
                )) for stage_result in results_batch
        ])
        
        for events_stage_name, events_group in events.items():  

            await logger.filesystem.aio['hedra.reporting'].debug(
                f'{metadata_string} - Events group - {events_group.events_group_id} - created for Stage - {events_stage_name} - Processed - {events_group.total}'
            )

            events_group.calculate_partial_group_stats()

        elapsed = time.monotonic() - start

        await logger.filesystem.aio['hedra.core'].info(f'{metadata_string} - Results aggregation complete - Took: {round(elapsed, 2)} seconds')

        return dill.dumps(events)


def process_results_batch(config: Dict[str, Any]):
    import asyncio
    import uvloop
    uvloop.install()
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    
    graph_name = config.get('graph_name')
    graph_id = config.get('graph_id')
    source_stage_name = config.get('source_stage_name')
    source_stage_id = config.get('source_stage_id')

    thread_id = threading.current_thread().ident
    process_id = os.getpid()

    meetadata_string = f'Graph - {graph_name}:{graph_id} - thread:{thread_id} - process:{process_id} - Stage: {source_stage_name}:{source_stage_id} - '

    stage_name = config.get('analyze_stage_name')
    custom_metric_hook_names = config.get('analyze_stage_metric_hooks', [])
    results_batch = config.get('analyze_stage_batched_results', [])

    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)

    try:
        return loop.run_until_complete(
            process_batch(
                stage_name,
                custom_metric_hook_names,
                results_batch,
                meetadata_string
            )
        )

    except Exception as e:
        raise e



import asyncio
import json
import re
import psutil
import uuid
from typing import List
from concurrent.futures import ThreadPoolExecutor
from hedra.logging import HedraLogger
from hedra.reporting.events.types.base_event import BaseEvent
from hedra.reporting.metric import MetricsSet

try:

    from google.cloud import storage
    from .google_cloud_storage_config import GoogleCloudStorageConfig
    has_connector = True

except Exception:
    storage = None
    GoogleCloudStorageConfig = None
    has_connector = False

class GoogleCloudStorage:

    def __init__(self, config: GoogleCloudStorageConfig) -> None:
        self.service_account_json_path = config.service_account_json_path
        self.bucket_namespace = config.bucket_namespace
        self.events_bucket_name = config.events_bucket
        self.metrics_bucket_name = config.metrics_bucket
        self.shared_metrics_bucket_name = 'stage_metrics'
        self.errors_bucket_name = 'stage_errors'

        self.credentials = None
        self.client = None

        self._events_bucket = None
        self._shared_metrics_bucket = None
        self._metrics_bucket = None
        self._errors_bucket = None
        self._custom_metrics_buckets = {}

        self.session_uuid = str(uuid.uuid4())
        self.metadata_string: str = None
        self.logger = HedraLogger()
        self.logger.initialize()

        self._executor = ThreadPoolExecutor(max_workers=psutil.cpu_count(logical=False))
        self._loop = asyncio.get_event_loop()

    async def connect(self):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Opening amd authorizing connection to Google Cloud - Loading account config from - {self.service_account_json_path}')
        self.client = storage.Client.from_service_account_json(self.service_account_json_path)

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Opened connection to Google Cloud - Loaded account config from - {self.service_account_json_path}')

    async def submit_events(self, events: List[BaseEvent]):

        try:

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating Events bucket at - Namespace: {self.bucket_namespace} - Bucket: {self.events_bucket_name} if not exists')

            self._events_bucket = await self._loop.run_in_executor(
                self._executor,
                self.client.get_bucket,
                f'{self.bucket_namespace}_{self.events_bucket_name}'
            )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created Events bucket at - Namespace: {self.bucket_namespace} - Bucket: {self.events_bucket_name}')
        
        except Exception:

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Setting Events bucket as - Namespace: {self.bucket_namespace} - Bucket: {self.events_bucket_name}')

            self._events_bucket = await self._loop.run_in_executor(
                self._executor,
                self.client.create_bucket,
                f'{self.bucket_namespace}_{self.events_bucket_name}'
            )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Set Events bucket as - Namespace: {self.bucket_namespace} - Bucket: {self.events_bucket_name}')

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Events to - Namespace: {self.bucket_namespace} - Bucket: {self.events_bucket_name}')
        for event in events:
            blob = await self._loop.run_in_executor(
                self._executor,
                self._events_bucket.blob,
                event.name
            )

            await self._loop.run_in_executor(
                self._executor,
                blob.upload_from_string,
                json.dumps(event.record)
            )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Events to - Namespace: {self.bucket_namespace} - Bucket: {self.events_bucket_name}')

    async def submit_common(self, metrics_sets: List[MetricsSet]):

        try:

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating Shared Metrics bucket at - Namespace: {self.bucket_namespace} - Bucket: {self.shared_metrics_bucket_name} if not exists')

            self._shared_metrics_bucket = await self._loop.run_in_executor(
                self._executor,
                self.client.get_bucket,
                f'{self.bucket_namespace}_{self.shared_metrics_bucket_name}'
            )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created Shared Metrics bucket at - Namespace: {self.bucket_namespace} - Bucket: {self.shared_metrics_bucket_name}')
        
        except Exception:

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Setting Shared Metrics bucket as - Namespace: {self.bucket_namespace} - Bucket: {self.shared_metrics_bucket_name}')

            self._shared_metrics_bucket = await self._loop.run_in_executor(
                self._executor,
                self.client.create_bucket,
                f'{self.bucket_namespace}_{self.shared_metrics_bucket_name}'
            )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Set Shared Metrics bucket as - Namespace: {self.bucket_namespace} - Bucket: {self.shared_metrics_bucket_name}')


        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Shared Metrics to - Namespace: {self.bucket_namespace} - Bucket: {self.shared_metrics_bucket_name}')
        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Shared Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            blob = await self._loop.run_in_executor(
                self._executor,
                self.metrics_bucket.blob,
                metrics_set.name
            )

            await self._loop.run_in_executor(
                self._executor,
                blob.upload_from_string,
                json.dumps({
                    'name': metrics_set.name,
                    'stage': metrics_set.stage,
                    'group': 'common',
                    **metrics_set.common_stats
                })
            )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Shared Metrics to - Namespace: {self.bucket_namespace} - Bucket: {self.shared_metrics_bucket_name}')

    async def submit_metrics(self, metrics: List[MetricsSet]):

        try:

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating Metrics bucket at - Namespace: {self.bucket_namespace} - Bucket: {self.metrics_bucket_name} if not exists')

            self.metrics_bucket = await self._loop.run_in_executor(
                self._executor,
                self.client.get_bucket,
                f'{self.bucket_namespace}_{self.metrics_bucket_name}'
            )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created Metrics bucket at - Namespace: {self.bucket_namespace} - Bucket: {self.metrics_bucket_name}')
        
        except Exception:

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Setting Metrics bucket as - Namespace: {self.bucket_namespace} - Bucket: {self.metrics_bucket_name}')

            self.metrics_bucket = await self._loop.run_in_executor(
                self._executor,
                self.client.create_bucket,
                f'{self.bucket_namespace}_{self.metrics_bucket_name}'
            )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Set Metrics bucket as - Namespace: {self.bucket_namespace} - Bucket: {self.metrics_bucket_name}')


        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Metrics to - Namespace: {self.bucket_namespace} - Bucket: {self.metrics_bucket_name}')
        for metrics_set in metrics:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Shared Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            for group_name, group in metrics_set.groups.items():
                blob = await self._loop.run_in_executor(
                    self._executor,
                    self.metrics_bucket.blob,
                    f'{metrics_set.name}_{group_name}'
                )

                await self._loop.run_in_executor(
                    self._executor,
                    blob.upload_from_string,
                    json.dumps({
                        **group.record,
                        'group': group_name
                    })
                )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Metrics to - Namespace: {self.bucket_namespace} - Bucket: {self.metrics_bucket_name}')

    async def submit_custom(self, metrics_sets: List[MetricsSet]):

        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Custom Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            for custom_group_name, group in metrics_set.custom_metrics.items():

                custom_bucket_name = f'{self.bucket_namespace}_{custom_group_name}_metrics'

                try:

                    await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating Custom Metrics bucket at - Namespace: {self.bucket_namespace} - Bucket: {custom_bucket_name} if not exists')

                    self._custom_metrics_buckets[custom_bucket_name] = await self._loop.run_in_executor(
                        self._executor,
                        self.client.get_bucket,
                        custom_bucket_name
                    )

                    await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created Custom Metrics bucket at - Namespace: {self.bucket_namespace} - Bucket: {custom_bucket_name}')
                
                except Exception:

                    await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Setting Custom Metrics bucket at - Namespace: {self.bucket_namespace} - Bucket: {custom_bucket_name}')

                    self._custom_metrics_buckets[custom_bucket_name] = await self._loop.run_in_executor(
                        self._executor,
                        self.client.create_bucket,
                        custom_bucket_name
                    )

                    await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Set Custom Metrics bucket at - Namespace: {self.bucket_namespace} - Bucket: {custom_bucket_name}')
                
                await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Custom Metrics to - Namespace: {self.bucket_namespace} - Bucket: {custom_bucket_name}')

                blob = await self._loop.run_in_executor(
                    self._executor,
                    self.metrics_bucket.blob,
                    f'{metrics_set.name}_{custom_group_name}'
                )

                await self._loop.run_in_executor(
                    self._executor,
                    blob.upload_from_string,
                    json.dumps({
                        'name': metrics_set.name,
                        'stage': metrics_set.stage,
                        'group': custom_group_name,
                        **group
                    })
                )

                await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Custom Metrics to - Namespace: {self.bucket_namespace} - Bucket: {custom_bucket_name}')

    async def submit_errors(self, metrics_sets: List[MetricsSet]):

        try:

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating Errors Metrics bucket at - Namespace: {self.bucket_namespace} - Bucket: {self.errors_bucket_name} if not exists')

            self._errors_bucket = await self._loop.run_in_executor(
                self._executor,
                self.client.get_bucket,
                f'{self.bucket_namespace}_errors'
            )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created Errors Metrics bucket at - Namespace: {self.bucket_namespace} - Bucket: {self.errors_bucket_name}')
        
        except Exception:

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Setting Error Metrics bucket as - Namespace: {self.bucket_namespace} - Bucket: {self.errors_bucket_name}')

            self._errors_bucket = await self._loop.run_in_executor(
                self._executor,
                self.client.create_bucket,
                f'{self.bucket_namespace}_errors'
            )
            
            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Set Errors Metrics bucket as - Namespace: {self.bucket_namespace} - Bucket: {self.errors_bucket_name}')


        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Error Metrics to - Namespace: {self.bucket_namespace} - Bucket: {self.errors_bucket_name}')
        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Errors Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            for error in metrics_set.errors:
                error_message = re.sub(
                    '[^0-9a-zA-Z]+', 
                    '_',
                    error.get(
                        'message'
                    ).lower()
                )

                
                blob = await self._loop.run_in_executor(
                    self._executor,
                    self.metrics_bucket.blob,
                    f'{metrics_set.name}_{error_message}'
                )

                await self._loop.run_in_executor(
                    self._executor,
                    blob.upload_from_string,
                    json.dumps({
                        'metric_name': metrics_set.name,
                        'metric_stage': metrics_set.stage,
                        'error_message': error.get('message'),
                        'error_count': error.get('count')
                    })
                )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Errors Metrics to - Namespace: {self.bucket_namespace} - Bucket: {self.errors_bucket_name}')

            

    async def close(self):
        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Closing session - {self.session_uuid}')
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Closing Google Cloud connection')
        await self._loop.run_in_executor(
            None,
            self.client.close
        )

        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Session Closed - {self.session_uuid}')
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Closed Google Cloud connection')


import asyncio
from typing import List, Union
from datetime import datetime
from hedra.core.graphs.events import Event
from hedra.core.graphs.hooks.registry.registry_types import (
    EventHook, 
    SaveHook,
    RestoreHook,
    ContextHook
)
from hedra.core.graphs.hooks.hook_types.internal import Internal
from hedra.core.graphs.hooks.hook_types.hook_type import HookType
from hedra.core.graphs.stages.types.stage_types import StageTypes
from hedra.core.graphs.stages.base.stage import Stage


class Checkpoint(Stage):
    stage_type=StageTypes.CHECKPOINT

    def __init__(self) -> None:
        super().__init__()
        self.previous_stage = None
        self.accepted_hook_types = [ 
            HookType.CONTEXT ,
            HookType.EVENT, 
            HookType.RESTORE,
            HookType.SAVE, 
        ]

        self.requires_shutdown = True

    @Internal()
    async def run(self):

        events: List[Union[EventHook, Event]] = [event for event in self.hooks[HookType.EVENT]]
        pre_events: List[EventHook] = [
            event for event in events if isinstance(event, EventHook) and event.pre
        ]
        
        if len(pre_events) > 0:
            pre_event_names = ", ".join([
                event.shortname for event in pre_events
            ])

            await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Executing PRE events - {pre_event_names}')
            await asyncio.wait([
                asyncio.create_task(event.call()) for event in pre_events
            ], timeout=self.stage_timeout)

        restore_hooks: List[RestoreHook] = self.hooks[HookType.RESTORE]
        restore_paths: List[str] = [restore_hook.restore_path for restore_hook in restore_hooks]
        restore_hooks_count = len(restore_hooks)

        if restore_hooks_count > 0:
            await self.logger.spinner.append_message(f'Executing Restore checkpoints for - {restore_hooks_count} - items')
            await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Executing Restore checkpoints for - {restore_hooks_count} - items')

        for restore_hook in restore_hooks:
            await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Restore Hook {restore_hook.name}:{restore_hook.hook_id} - Restoring data from checkpoint file - {restore_hook.restore_path}')

        await asyncio.gather(*[
            asyncio.create_task(
                restore_hook.call(self.context)
            ) for restore_hook in restore_hooks
        ])

        save_hooks: List[SaveHook] = self.hooks[HookType.SAVE]
        allowed_save_hooks = [save_hook for save_hook in save_hooks if save_hook.save_path not in restore_paths]
        save_hooks_count = len(allowed_save_hooks)
        if save_hooks_count:
            await self.logger.spinner.append_message(f'Executing Save checkpoints for - {save_hooks_count} - items')
            await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Executing Save checkpoints for - {len(save_hooks)} - items')

        
        for save_hook in allowed_save_hooks:
            await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Save Hook {save_hook.name}:{save_hook.hook_id} - Saving data to checkpoint file - {save_hook.save_path}')
        
        await asyncio.gather(*[
            asyncio.create_task(
                save_hook.call(self.context)
            ) for save_hook in allowed_save_hooks
        ])
        
        post_events: List[EventHook] = [
            event for event in events if isinstance(event, EventHook) and event.pre is False
        ]

        if len(post_events) > 0:
            post_event_names = ", ".join([
                event.shortname for event in post_events
            ])

            await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Executing POST events - {post_event_names}')
            await asyncio.wait([
                asyncio.create_task(event.call()) for event in post_events
            ], timeout=self.stage_timeout)

        context_hooks: List[ContextHook] = self.hooks[HookType.CONTEXT]
        await asyncio.gather(*[
            asyncio.create_task(context_hook.call(self.context)) for context_hook in context_hooks
        ])

        await self.logger.spinner.set_default_message(f"Checkpoint complete - Restored {restore_hooks_count} items and saved {save_hooks_count} items")
        await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Completed checkpoints for - {len(save_hooks)} - items')




import asyncio
import os
import uuid

from .worker_server import WorkerServer


class WorkerManager:

    def __init__(self, config, reporter_config) -> None:
        logger = Logger()
        
        self.worker_type = config.runner_mode
        self.ephemeral_job_name = 'ephemeral-job'
  
        self.session_logger = logger.generate_logger()
        self.worker_id = uuid.uuid4()

        self.worker_ip = config.distributed_config.get(
            'worker_ip',
            os.getenv('WORKER_IP', 'localhost')
        )
        self.worker_port = config.distributed_config.get(
            'worker_port',
            os.getenv('WORKER_PORT', 6670)
        )
        self.worker_server_port = config.distributed_config.get(
            'worker_server_port',
            os.getenv('WORKER_SERVER_PORT',  6671)
        )

        self.worker_address = f'{self.worker_ip}:{self.worker_port}'
        self._live_progress_updates = config.distributed_config.get('live_progress_updates', False)
        
        self.session_logger.info(f'Initializing worker at - {self.worker_address}')


        self.server = WorkerServer(config, reporter_config, self.worker_id)
        self.running = True

    async def register(self):
        self.session_logger.info(f'\nStarting worker at - {self.worker_ip}:{self.worker_server_port}.')
        self.session_logger.info(f'Waiting for - {self.server.service.job_registry.leaders.registry.initial_count} - leaders.')
        await self.server.register_worker()
        await self.server.service.job_registry.wait_for_leaders()

        self.session_logger.info(f'\nStarting monitor...')
        if self.server.service.job_registry.monitoring is False:
            await self.server.start_monitor()
        
        self.session_logger.info(f'Startup complete!')

    async def run(self):
        self.session_logger.debug('Live updates? - {updates}'.format(
            updates=self._live_progress_updates
        ))
        
        await self.start_server()

    async def start_server(self):
        self.session_logger.debug('Running worker server...')
        await self.server.run()


    async def wait(self):
        while self.running:

            new_leader_queue = self.server.service.new_leader_queue
            not_empty = await new_leader_queue.not_empty()

            if not_empty:
                await self.server.add_new_leaders(new_leader_queue)
                await new_leader_queue.clear()
                self.server.service.new_leader_queue = new_leader_queue
            
            await self.server.service.job_registry.leaders.check_leader_heartbeat()
            await self.server.service.job_registry.update_job_leaders()

            if self.worker_type == 'ephemeral-worker':
                job = self.server.service.job_registry[self.ephemeral_job_name]
                if job:
                    if job.status == 'completed' or job.status == 'failed':
                        self.running = False
                        return

            await asyncio.sleep(1)

    async def stop(self):
        await self.server.stop_server()




import random
import time
import asyncio
import uuid
import psutil
import math
from typing import Dict, List, Union
from hedra.core.personas.types.default_persona import DefaultPersona
from hedra.core.graphs.hooks.registry.registry_types import ActionHook, TaskHook
from hedra.core.graphs.hooks.hook_types.hook_type import HookType
from hedra.core.engines.client.config import Config
from hedra.core.personas.types.types import PersonaTypes


class WeightedSelectionPersona(DefaultPersona):
    
    __slots__ = (
        'weights',
        'indexes',
        'sample'
    )

    def __init__(self, config: Config):
        super().__init__(config)

        self.persona_id = str(uuid.uuid4())
        self.weights: List[int] = []
        self.indexes: List[int] = []
        self.sample: List[int] = []
        self.type = PersonaTypes.WEIGHTED
        
    @classmethod
    def about(cls):
        return '''
        Weighted Persona - (sequence)

        Executes batches of actions of the batch size specified by the --batch-size  argument. Actions for each batch are resampled each iteration according 
        to their specified "weight". As with other personas, the Weighted persona will execute for the total amount of time specified by the --total-time 
        argument. You may specify a wait between batches (between each step) by specifying an integer number of seconds via the --batch-interval argument.
        '''

    def setup(self, hooks: Dict[HookType, List[Union[ActionHook, TaskHook]]], metadata_string: str):

        self.metadata_string = f'{metadata_string} Persona: {self.type.capitalize()}:{self.persona_id} - '
        
        actions = hooks.get(HookType.ACTION)
        actions.extend(
            hooks.get(HookType.TASK)
        )
        self.actions_count = len(actions)

        self.weights = [action.metadata.weight for action in actions]
        self.indexes = [idx for idx in range(self.actions_count)]

        self.sample = random.choices(
            self.indexes,
            weights=self.weights,
            k=self.batch.size
        )

        self._hooks = actions


    async def generator(self, total_time):
        elapsed = 0
        max_pool_size = math.ceil(self.batch.size * (psutil.cpu_count(logical=False) * 2)/self.workers)

        start = time.time()
        while elapsed < total_time:
            for action_idx in self.sample:
                yield action_idx
                
                await asyncio.sleep(0)
                elapsed = time.time() - start
            
                if self._hooks[action_idx].session.active > max_pool_size:
                    try:
                        max_wait = total_time - elapsed
                        await asyncio.wait_for(
                            self._hooks[action_idx].session.wait_for_active_threshold(),
                            timeout=max_wait
                        )
                    except asyncio.TimeoutError:
                        pass

            self.sample = random.choices(
                self.indexes,
                weights=self.weights,
                k=self.batch.size
            )



import json
import dill
import threading
import os
import importlib
import inspect
import ntpath
import sys
from collections import defaultdict
from typing import Any, Dict, List, Union
from pathlib import Path
from hedra.core.engines.client.config import Config
from hedra.core.graphs.hooks.registry.registrar import registrar
from hedra.core.graphs.stages.optimize.optimization import Optimizer
from hedra.core.graphs.stages.base.stage import Stage
from hedra.core.graphs.stages.setup.setup import Setup
from hedra.core.engines.types.registry import registered_engines
from hedra.core.personas.persona_registry import registered_personas
from hedra.plugins.types.plugin_types import PluginType
from hedra.plugins.types.engine.engine_plugin import EnginePlugin
from hedra.plugins.types.persona.persona_plugin import PersonaPlugin
from hedra.core.graphs.stages.execute import Execute
from hedra.core.graphs.stages.base.import_tools import import_stages, import_plugins
from hedra.core.graphs.stages.optimize.optimization.algorithms import registered_algorithms
from hedra.logging import (
    HedraLogger,
    LoggerTypes,
    logging_manager
)
from hedra.core.personas import get_persona




def optimize_stage(serialized_config: str):

    import asyncio
    import uvloop
    uvloop.install()
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)

    hedra_config_filepath = os.path.join(
        os.getcwd(),
        '.hedra.json'
    )

    hedra_config = {}
    if os.path.exists(hedra_config_filepath):
        with open(hedra_config_filepath, 'r') as hedra_config_file:
            hedra_config = json.load(hedra_config_file)

    logging_config = hedra_config.get('logging', {})
    logfiles_directory = logging_config.get(
        'logfiles_directory',
        os.getcwd()
    )

    log_level = logging_config.get('log_level', 'info')

    logging_manager.disable(
        LoggerTypes.DISTRIBUTED,
        LoggerTypes.DISTRIBUTED_FILESYSTEM
    )

    logging_manager.update_log_level(log_level)
    logging_manager.logfiles_directory = logfiles_directory


    logger = HedraLogger()
    logger.initialize()
    logger.filesystem.sync.create_logfile('hedra.core.log')
    logger.filesystem.sync.create_logfile('hedra.optimize.log')

    logger.filesystem.create_filelogger('hedra.core.log')
    logger.filesystem.create_filelogger('hedra.optimize.log')

    thread_id = threading.current_thread().ident
    process_id = os.getpid()

    optimization_config: Dict[str, Union[str, int, Any]] = dill.loads(serialized_config)

    graph_name: str = optimization_config.get('graph_name')
    graph_path: str= optimization_config.get('graph_path')
    graph_id: str = optimization_config.get('graph_id')
    source_stage_name: str = optimization_config.get('source_stage_name')
    source_stage_id: str = optimization_config.get('source_stage_id')
    execute_stage_name: str = optimization_config.get('execute_stage_name')
    execute_stage_config: Config = optimization_config.get('execute_stage_config')
    execute_stage_plugins: Dict[PluginType, List[str]] = optimization_config.get('execute_stage_plugins')
    optimize_params: List[str] = optimization_config.get('optimize_params')
    optimize_iterations: int = optimization_config.get('optimizer_iterations')
    optimizer_algorithm: str = optimization_config.get('optimizer_algorithm')
    time_limit: int = optimization_config.get('time_limit')
    batch_size: int = optimization_config.get('execute_stage_batch_size')

    metadata_string = f'Graph - {graph_name}:{graph_id} - thread:{thread_id} - process:{process_id} - Stage: {source_stage_name}:{source_stage_id} - '

    discovered = import_stages(graph_path)

    execute_stage: Stage = discovered.get(execute_stage_name)()


    execute_stage_actions = defaultdict(list)
    for hook_name in optimization_config.get('execute_stage_hooks'):
        hook = registrar.all.get(hook_name)

        hook._call = hook._call.__get__(execute_stage, execute_stage.__class__)
        setattr(execute_stage, hook.shortname, hook._call)

        if inspect.ismethod(hook.call) is False:
            hook.call = hook.call.__get__(execute_stage, execute_stage.__class__)
            setattr(execute_stage, hook.shortname, hook.call)

        execute_stage_actions[hook.hook_type].append(hook)

    execute_stage.hooks.update(execute_stage_actions)
    execute_stage_config.batch_size = batch_size

    plugins_by_type = import_plugins(graph_path)

    stage_persona_plugins: List[str] = execute_stage_plugins[PluginType.PERSONA]
    persona_plugins: Dict[str, PersonaPlugin] = plugins_by_type[PluginType.PERSONA]

    stage_engine_plugins: List[str] = execute_stage_plugins[PluginType.ENGINE]
    engine_plugins: Dict[str, EnginePlugin] = plugins_by_type[PluginType.ENGINE]
    
    for plugin_name in stage_persona_plugins:
        plugin = persona_plugins.get(plugin_name)
        plugin.name = plugin_name
        registered_personas[plugin_name] = lambda config: plugin(config)
    
    for plugin_name in stage_engine_plugins:
        plugin = engine_plugins.get(plugin_name)
        plugin.name = plugin_name
        registered_engines[plugin_name] = lambda config: plugin(config)

    for plugin_name, plugin in plugins_by_type[PluginType.OPTIMIZER].items():
        registered_algorithms[plugin_name] = plugin

    setup_stage = Setup()
    setup_stage.plugins_by_type = plugins_by_type
    setup_stage.generation_setup_candidates = 1
    setup_stage.stages[execute_stage.name] = execute_stage
    setup_stage.config = execute_stage_config

    stages = loop.run_until_complete(setup_stage.run())
    
    setup_execute_stage: Execute = stages.get(execute_stage_name)

    logger.filesystem.sync['hedra.optimize'].info(f'{metadata_string} - Setting up Optimization')

    optimizer = Optimizer({
        'graph_name': graph_name,
        'graph_id': graph_id,
        'source_stage_name': source_stage_name,
        'source_stage_id': source_stage_id,
        'params': optimize_params,
        'stage_name': execute_stage_name,
        'stage_config': execute_stage_config,
        'stage_hooks': setup_execute_stage.hooks,
        'iterations': optimize_iterations,
        'algorithm': optimizer_algorithm,
        'time_limit': time_limit
    })

    results = optimizer.optimize(loop)

    execute_stage_config.batch_size = results.get('optimized_batch_size', execute_stage_config.batch_size)
    execute_stage_config.batch_interval = results.get('optimized_batch_interval', execute_stage_config.batch_gradient)
    execute_stage_config.batch_gradient = results.get('optimized_batch_gradient', execute_stage_config.batch_gradient)

    logger.filesystem.sync['hedra.optimize'].info(f'{optimizer.metadata_string} - Optimization complete')

    return {
        'stage': execute_stage.name,
        'config': execute_stage_config,
        'params': results
    }


from .bootstrap_pb2 import (
    ServiceRegistrationStatusRequest,
    ServiceRegistrationStatusResponse,
    ServiceRegisterRequest
)


from .bootstrap_pb2_grpc import (
    BootstrapServerServicer,
    BootstrapServerStub,
    add_BootstrapServerServicer_to_server
)


from pydantic import BaseModel
from hedra.reporting.types.common.types import ReporterTypes


class S3Config(BaseModel):
    aws_access_key_id: str
    aws_secret_access_key: str
    region_name: str
    buckets_namespace: str
    events_bucket: str='events'
    metrics_bucket: str='metrics'
    reporter_type: ReporterTypes=ReporterTypes.S3


from typing import Any, Dict, Callable
from scipy.optimize import OptimizeResult
from hedra.plugins.types.optimizer import (
    OptimizerPlugin,
    get,
    optimize,
    update
)


class CustomOptimizer(OptimizerPlugin):

    def __init__(self, config: Dict[str, Any]) -> None:
        super().__init__(config)

    @get()
    def get_params(self):
        return super().get_params()

    @update()
    def update_params(self):
        return super().update_params()

    @optimize()
    def run_dual_annealing(self, func: Callable[..., OptimizeResult]):
        pass


import math
import time
import asyncio
import uuid
import psutil
from asyncio import Task
from hedra.core.engines.client.config import Config
from hedra.core.personas.types.default_persona.default_persona import DefaultPersona
from hedra.core.personas.types.types import PersonaTypes
from .completed_counter import CompletedCounter


async def cancel_pending(pend: Task):
    try:
        pend.cancel()
        await pend
    
    except asyncio.CancelledError:
        pass


class ConstantArrivalPersona(DefaultPersona):

    __slots__ = (
        'completed_counter'
    )

    def __init__(self, config: Config):
        super(ConstantArrivalPersona, self).__init__(config)

        self.persona_id = str(uuid.uuid4())

        self.completed_counter = CompletedCounter()
        self.type = PersonaTypes.CONSTANT_ARRIVAL
            
    async def execute(self):

        hook_names = ', '.join([
            hook.name for hook in self._hooks
        ])

        await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Executing {self.actions_count} Hooks: {hook_names}')

        total_time = self.total_time

        await self.logger.filesystem.aio['hedra.core'].debug(f'{self.metadata_string} - Executing for a total of - {total_time} - seconds')

        await self.start_updates()

        self.start = time.monotonic()
        completed, pending = await asyncio.wait([
            asyncio.create_task(
                self.completed_counter.execute_action(
                    self._hooks[action_idx]
                )
            ) async for action_idx in self.generator(total_time)
        ], timeout=1)

        self.end = time.monotonic()
        self.pending_actions = len(pending)

        await self.logger.filesystem.aio['hedra.core'].debug(
            f'{self.metadata_string} - Execution completed with - {self.pending_actions} - actions left pending'
        )

        results = await asyncio.gather(*completed)
        
        await self.stop_updates()

        cleanup_start = time.monotonic()

        await asyncio.gather(*[
            asyncio.create_task(
                cancel_pending(pend)
            ) for pend in pending
        ], return_exceptions=True)

        cleanup_elapsed = time.monotonic() - cleanup_start
        await self.logger.filesystem.aio['hedra.core'].info(
            f'{self.metadata_string} - Cleanup completed - Resolved {self.pending_actions} pending actions in {round(cleanup_elapsed, 2)} seconds'
        )

        for hook in self._hooks:

            session_closed_start = time.monotonic()

            await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Closing session - {hook.session.session_id} - for Hook - {hook.name}:{hook.hook_id}')
            await hook.session.close()

            session_closed_elapsed = time.monotonic() - session_closed_start

            await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Closed session - {hook.session.session_id} - for Hook - {hook.name}:{hook.hook_id}. Took: {round(session_closed_elapsed, 2)} seconds')
  
        self.total_actions = len(set(results))
        self.total_elapsed = self.end - self.start
        self.optimized_params = None

        await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Completed execution')

        return results

    async def generator(self, total_time):
        elapsed = 0
        idx = 0
        action_idx = 0
        max_pool_size = math.ceil(self.batch.size * (psutil.cpu_count(logical=False) * 2)/self.workers)
        self.completed_counter.last_batch_size = self.batch.size

        start = time.time()
        while elapsed < total_time:
            yield action_idx
            
            await asyncio.sleep(0)
            elapsed = time.time() - start
            idx += 1

            if idx%self._hooks[action_idx].session.pool.size == 0:

                if self.completed_counter.completed_count  > 0:

                    if self.completed_counter.completed_count < self.batch.size:
                        increase_percentage = (self.batch.size - self.completed_counter.completed_count)/self.batch.size
                        increase_amount = math.ceil(increase_percentage * self.completed_counter.last_batch_size)

                        self.completed_counter.last_completed = self.completed_counter.completed_count
                        self.completed_counter.last_batch_size = self.completed_counter.last_batch_size + increase_amount

                        self._hooks[action_idx].session.extend_pool(increase_amount)
                        await asyncio.sleep(0)

                    elif self.completed_counter.completed_count > self.batch.size:
                        decrease_percentage = (self.completed_counter.completed_count - self.batch.size)/self.completed_counter.completed_count
                        decrease_amount = math.ceil(decrease_percentage * self.completed_counter.last_batch_size)

                        self.completed_counter.last_completed = self.completed_counter.completed_count
                        self.completed_counter.last_batch_size = self.completed_counter.last_batch_size - decrease_amount

                        self._hooks[action_idx].session.shrink_pool(decrease_amount)
                        await asyncio.sleep(0)
            
                    self.completed_counter.completed_count = 0

                await asyncio.sleep(self.batch.interval)

            action_idx = (action_idx + 1) % self.actions_count
            if self._hooks[action_idx].session.active > max_pool_size:
                    try:
                        max_wait = total_time - elapsed
                        await asyncio.wait_for(
                            self._hooks[action_idx].session.wait_for_active_threshold(),
                            timeout=max_wait
                        )
                    except asyncio.TimeoutError:
                        pass




from typing import Any
from hedra.core.graphs.stages.base.stage import Stage
from hedra.core.graphs.transitions.exceptions import InvalidTransitionError
from hedra.core.graphs.stages.types.stage_types import StageTypes
from hedra.logging import HedraLogger


async def idle_transition(current_stage: Stage, next_stage: Stage):

    logger = HedraLogger()
    logger.initialize()

    await logger.spinner.system.debug(f'{current_stage.metadata_string} - NoOp transition from {current_stage.name} to {next_stage.name}')
    await logger.filesystem.aio['hedra.core'].debug(f'{current_stage.metadata_string} - NoOp transition from {current_stage.name} to {next_stage.name}')

    next_stage.context = current_stage.context

    return None, StageTypes.IDLE


async def invalid_transition(current_stage: Stage, next_stage: Stage):
    return InvalidTransitionError(current_stage, next_stage), StageTypes.ERROR


async def exit_transition(current_stage: Stage, next_stage: Stage):

    logger = HedraLogger()
    logger.initialize()

    await logger.spinner.system.debug(f'{current_stage.metadata_string} - Executing transition from {current_stage.name} to {next_stage.name}')
    await logger.filesystem.aio['hedra.core'].debug(f'{current_stage.metadata_string} - Executing transition from {current_stage.name} to {next_stage.name}')

    await current_stage.run()
    next_stage.context = None
    current_stage.context = None

    await logger.spinner.system.debug(f'{current_stage.metadata_string} - Completed transition from {current_stage.name} to {next_stage.name}')
    await logger.filesystem.aio['hedra.core'].debug(f'{current_stage.metadata_string} - Completed transition from {current_stage.name} to {next_stage.name}')

    return None, None

async def error_transition(current_stage: Stage, next_stage: Stage):

    logger = HedraLogger()
    logger.initialize()

    await logger.spinner.system.debug(f'{current_stage.metadata_string} - Executing transition from {current_stage.name} to {next_stage.name}')
    await logger.filesystem.aio['hedra.core'].debug(f'{current_stage.metadata_string} - Executing transition from {current_stage.name} to {next_stage.name}')

    await next_stage.run()
    next_stage.context = None
    current_stage.context = None

    await logger.spinner.system.debug(f'{current_stage.metadata_string} - Completed transition from {current_stage.name} to {next_stage.name}')
    await logger.filesystem.aio['hedra.core'].debug(f'{current_stage.metadata_string} - Completed transition from {current_stage.name} to {next_stage.name}')

    return None, None


from datetime import datetime
import uuid
from typing import List

from numpy import float32, float64, int16, int32, int64
from hedra.reporting.events.types.base_event import BaseEvent
from hedra.reporting.metric import MetricsSet


try:
    import sqlalchemy
    from sqlalchemy.schema import CreateTable
    from sqlalchemy.sql import func
    from sqlalchemy.dialects.postgresql import UUID
    from hedra.reporting.types.postgres.postgres import Postgres
    from .timescaledb_config import TimescaleDBConfig
    has_connector=True

except Exception:
    sqlalchemy = None
    UUID = None
    from hedra.reporting.types.empty import Empty as Postgres
    CreateTable = None
    TimescaleDBConfig = None
    has_connector = False


class TimescaleDB(Postgres):

    def __init__(self, config: TimescaleDBConfig) -> None:
        super().__init__(config)

    async def submit_events(self, events: List[BaseEvent]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Events to Table - {self.events_table_name}')

        async with self._connection.begin() as transaction:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Events to Table - {self.events_table_name} - Initiating transaction')
        
            for event in events:

                if self._events_table is None:
                    await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Events table - {self.events_table_name} - if not exists')

                    events_table = sqlalchemy.Table(
                        self.events_table_name,
                        self.metadata,
                        sqlalchemy.Column('id', UUID(as_uuid=True), primary_key=True, default=uuid.uuid4),
                        sqlalchemy.Column('name', sqlalchemy.VARCHAR(255)),
                        sqlalchemy.Column('stage', sqlalchemy.VARCHAR(255)),
                        sqlalchemy.Column('request_time', sqlalchemy.Float),
                        sqlalchemy.Column('succeeded', sqlalchemy.Boolean),
                        sqlalchemy.Column('time', sqlalchemy.TIMESTAMP(timezone=False), nullable=False, default=datetime.now())
                    )

                    await self._connection.execute(CreateTable(events_table, if_not_exists=True))
                    await self._connection.execute(
                        f"SELECT create_hypertable('{self.events_table_name}', 'time', migrate_data => true, if_not_exists => TRUE, create_default_indexes=>FALSE);"
                    )

                    await self._connection.execute(f"CREATE INDEX ON {self.events_table_name} (name, time DESC);")

                    self._events_table = events_table
                    await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Created or set Events table - {self.events_table_name}')
                
                record = event.record
                record['request_time'] = record['time']
                del record['time']

                await self._connection.execute(self._events_table.insert(values={
                    **record,
                    'time': datetime.now().timestamp()
                }))
                
            await transaction.commit()
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Events to Table - {self.events_table_name} - Transaction committed')

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Events to Table - {self.events_table_name}')

    async def submit_common(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Shared Metrics to Table - {self.shared_metrics_table_name}')
        
        async with self._connection.begin() as transaction:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Shared Metrics to Table - {self.shared_metrics_table_name} - Initiating transaction')

            if self._stage_metrics_table is None:
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Shared Metrics table - {self.shared_metrics_table_name} - if not exists')

                stage_metrics_table = sqlalchemy.Table(
                    self.stage_metrics_table_name,
                    self.metadata,
                    sqlalchemy.Column('id', UUID(as_uuid=True), default=uuid.uuid4),
                    sqlalchemy.Column('name', sqlalchemy.VARCHAR(255)),
                    sqlalchemy.Column('stage', sqlalchemy.VARCHAR(255)),
                    sqlalchemy.Column('group', sqlalchemy.VARCHAR(255)),
                    sqlalchemy.Column('total', sqlalchemy.BIGINT),
                    sqlalchemy.Column('succeeded', sqlalchemy.BIGINT),
                    sqlalchemy.Column('failed', sqlalchemy.BIGINT),
                    sqlalchemy.Column('actions_per_second', sqlalchemy.FLOAT),
                    sqlalchemy.Column('time', sqlalchemy.TIMESTAMP(timezone=False), nullable=False, default=datetime.now())
                )

                await self._connection.execute(CreateTable(stage_metrics_table, if_not_exists=True))
                await self._connection.execute(
                    f"SELECT create_hypertable('{self.stage_metrics_table_name}', 'time', migrate_data => true, if_not_exists => TRUE, create_default_indexes=>FALSE);"
                )

                await self._connection.execute(f"CREATE INDEX ON {self.stage_metrics_table_name} (name, time DESC);")

                self._stage_metrics_table = stage_metrics_table

                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Created or set Shared Metrics table - {self.shared_metrics_table_name}')

            for metrics_set in metrics_sets:
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Shared Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')
                await self._connection.execute(
                    self._stage_metrics_table.insert(values={
                        'name': metrics_set.name,
                        'stage': metrics_set.stage,
                        'group': 'common',
                        **metrics_set.common_stats
                    })
                )

            await transaction.commit()
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Shared Metrics to Table - {self.shared_metrics_table_name} - Transaction committed')

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Shared Metrics to Table - {self.shared_metrics_table_name}')

    async def submit_metrics(self, metrics: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Metrics to Table - {self.metrics_table_name}')

        async with self._connection.begin() as transaction:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Metrics to Table - {self.metrics_table_name} - Initiating transaction')

            for metrics_set in metrics:
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

                if self._metrics_table is None:
                    await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Metrics table - {self.metrics_table_name} - if not exists')

                    metrics_table = sqlalchemy.Table(
                        self.metrics_table_name,
                        self.metadata,
                        sqlalchemy.Column('id', UUID(as_uuid=True), default=uuid.uuid4),
                        sqlalchemy.Column('name', sqlalchemy.VARCHAR(255)),
                        sqlalchemy.Column('stage', sqlalchemy.VARCHAR(255)),
                        sqlalchemy.Column('group', sqlalchemy.TEXT),
                        sqlalchemy.Column('median', sqlalchemy.FLOAT),
                        sqlalchemy.Column('mean', sqlalchemy.FLOAT),
                        sqlalchemy.Column('variance', sqlalchemy.FLOAT),
                        sqlalchemy.Column('stdev', sqlalchemy.FLOAT),
                        sqlalchemy.Column('minimum', sqlalchemy.FLOAT),
                        sqlalchemy.Column('maximum', sqlalchemy.FLOAT),
                        sqlalchemy.Column('time', sqlalchemy.TIMESTAMP(timezone=False), nullable=False, default=datetime.now())
                    )

                    for quantile in metrics_set.quantiles:
                        metrics_table.append_column(
                            sqlalchemy.Column(f'{quantile}', sqlalchemy.FLOAT)
                        )

                    for custom_field_name, sql_alchemy_type in metrics_set.custom_schemas:
                        metrics_table.append_column(custom_field_name, sql_alchemy_type)

                    await self._connection.execute(CreateTable(metrics_table, if_not_exists=True))
                    await self._connection.execute(
                        f"SELECT create_hypertable('{self.metrics_table_name}', 'time', migrate_data => true, if_not_exists => TRUE, create_default_indexes=>FALSE);"
                    )

                    await self._connection.execute(f"CREATE INDEX ON {self.metrics_table_name} (name, time DESC);")

                    self._metrics_table = metrics_table

                    await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Created or set Metrics table - {self.metrics_table_name}')

                for group_name, group in metrics_set.groups.items():
                    await self._connection.execute(self._metrics_table.insert(values={
                        **group.record,
                        'group': group_name
                    }))

            await transaction.commit()
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Metrics to Table - {self.metrics_table_name} - Transaction committed')

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Metrics to Table - {self.metrics_table_name}')

    async def submit_custom(self, metrics_sets: List[MetricsSet]):

        async with self._connection.begin() as transaction:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Custom Metrics - Initiating transaction')

            for metrics_set in metrics_sets:
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Custom Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')
                
                for custom_group_name, group in metrics_set.custom_metrics.items():
                    custom_table_name = f'{custom_group_name}_metrics'
                    await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Custom Metrics to table - {custom_group_name}')

                    if self._custom_metrics_tables.get(custom_table_name) is None:
                        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Custom Metrics table - {custom_group_name} - if not exists')

                        custom_metrics_table = sqlalchemy.Table(
                            custom_table_name,
                            self.metadata,
                            sqlalchemy.Column('id', UUID(as_uuid=True), default=uuid.uuid4),
                            sqlalchemy.Column('name', sqlalchemy.VARCHAR(255)),
                            sqlalchemy.Column('stage', sqlalchemy.VARCHAR(255)),
                            sqlalchemy.Column('group', sqlalchemy.TEXT),
                            sqlalchemy.Column('time', sqlalchemy.TIMESTAMP(timezone=False), nullable=False, default=datetime.now())
                        )

                        for field, value in group.items():

                            if isinstance(value, (int, int16, int32, int64)):
                                custom_metrics_table.append_column(
                                    sqlalchemy.Column(field, sqlalchemy.INTEGER)
                                )

                            elif isinstance(value, (float, float32, float64)):
                                custom_metrics_table.append_column(
                                    sqlalchemy.Column(field, sqlalchemy.FLOAT)
                                )


                        await self._connection.execute(CreateTable(custom_metrics_table, if_not_exists=True))
                        await self._connection.execute(
                            f"SELECT create_hypertable('{custom_table_name}', 'time', migrate_data => true, if_not_exists => TRUE, create_default_indexes=>FALSE);"
                        )

                        await self._connection.execute(f"CREATE INDEX ON {custom_table_name} (name, time DESC);")

                        self._custom_metrics_tables[custom_group_name] = custom_metrics_table

                        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Created or set Custom Metrics table - {custom_group_name}')

                    await self._connection.execute(
                        self._custom_metrics_tables[custom_table_name].insert(values={
                            'name': metrics_set.name,
                            'stage': metrics_set.stage,
                            'group': custom_group_name,
                            **group
                        })
                    )

                    await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Custom Metrics to table - {custom_group_name}')

            await transaction.commit()
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Custom Metrics - Transaction committed')

    async def submit_errors(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Error Metrics to Table - {self.errors_table_name}')

        async with self._connection.begin() as transaction:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Error Metrics to Table - {self.errors_table_name} - Initiating transaction')

            for metrics_set in metrics_sets:
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Error Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

                if self._errors_table is None:
                    await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Error Metrics table - {self.errors_table_name} - if not exists')

                    errors_table = sqlalchemy.Table(
                        self.errors_table_name,
                        self.metadata,
                        sqlalchemy.Column('id', UUID(as_uuid=True), default=uuid.uuid4),
                        sqlalchemy.Column('metric_name', sqlalchemy.VARCHAR(255)),
                        sqlalchemy.Column('metrics_stage', sqlalchemy.TEXT),
                        sqlalchemy.Column('error_message', sqlalchemy.TEXT),
                        sqlalchemy.Column('error_count', sqlalchemy.BIGINT),
                        sqlalchemy.Column('time', sqlalchemy.TIMESTAMP(timezone=False), nullable=False, default=datetime.now())
                    )    

                    
                    await self._connection.execute(CreateTable(errors_table, if_not_exists=True))
                    await self._connection.execute(
                        f"SELECT create_hypertable('{self.errors_table_name}', 'time', migrate_data => true, if_not_exists => TRUE, create_default_indexes=>FALSE);"
                    )

                    await self._connection.execute(f"CREATE INDEX ON {self.errors_table_name}_errors (name, time DESC);")

                    self._errors_table = errors_table    

                    await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Created or set Error Metrics table - {self.errors_table_name}')

                for error in metrics_set.errors:
                    await self._connection.execute(self._metrics_table.insert(values={
                        'metric_name': metrics_set.name,
                        'metrics_stage': metrics_set.stage,
                        'error_message': error.get('message'),
                        'error_count': error.get('count')
                    }))
            
            await transaction.commit()
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Error Metrics to Table - {self.errors_table_name} - Transaction committed')

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Error Metrics to Table - {self.errors_table_name}')
                
    async def close(self):
        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Closing session - {self.session_uuid}')
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Closing connectiion to {self.sql_type} at - {self.host}')

        await self._connection.close()
        self._engine.terminate()
        await self._engine.wait_closed()

        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Session Closed - {self.session_uuid}')
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Closed connectiion to {self.sql_type} at - {self.host}')




from types import SimpleNamespace
from typing import Dict
from pydantic import BaseModel
from hedra.reporting.types.common.types import ReporterTypes


try:

    from google.cloud import bigquery
    has_connector = True

except Exception:
    bigquery = SimpleNamespace(SchemaField=None)
    has_connector = False


class BigQueryConfig(BaseModel):
    service_account_json_path: str
    project_name: str
    dataset_name: str
    events_table: str = 'events'
    metrics_table: str = 'metrics'
    retry_timeout: int = 10
    custom_fields: Dict[str, bigquery.SchemaField] = {}
    reporter_type: ReporterTypes=ReporterTypes.BigQuery

    class Config:
        arbitrary_types_allowed = True



import asyncio
import time
import uuid
from typing import Awaitable, Dict, Set, Tuple
from hedra.core.engines.types.common.timeouts import Timeouts
from hedra.core.engines.types.http2.pipe import HTTP2Pipe
from hedra.core.engines.types.http2.connection import HTTP2Connection
from hedra.core.engines.types.common.ssl import get_http2_ssl_context
from hedra.core.engines.types.common.concurrency import (
    Semaphore
)
from .pool import HTTP2Pool
from .action import HTTP2Action
from .result import HTTP2Result


HTTP2ResponseFuture = Awaitable[HTTP2Action]
HTTP2BatchResponseFuture = Awaitable[Tuple[Set[HTTP2ResponseFuture], Set[HTTP2ResponseFuture]]]


class MercuryHTTP2Client:

    __slots__ = (
        'session_id',
        'timeouts',
        '_hosts',
        'registered',
        'closed',
        'sem',
        'pool',
        'active',
        'waiter',
        'ssl_context'
    )

    def __init__(self, concurrency: int = 10**3, timeouts: Timeouts = Timeouts(), reset_connections: bool=False) -> None:

        self.session_id = str(uuid.uuid4())
        self.timeouts = timeouts

        self._hosts = {}
        self.registered: Dict[str, HTTP2Action] = {}
        self.closed = False
        
        self.sem = Semaphore(value=concurrency)
        self.pool: HTTP2Pool = HTTP2Pool(concurrency, self.timeouts, reset_connections=reset_connections)
        self.pool.create_pool()
        self.active = 0
        self.waiter = None

        self.ssl_context = get_http2_ssl_context()
        

    async def wait_for_active_threshold(self):
        if self.waiter is None:
            self.waiter = asyncio.get_event_loop().create_future()
            await self.waiter

    async def prepare(self, request: HTTP2Action) -> Awaitable[None]:
        try:
            request.ssl_context = self.ssl_context

            if self._hosts.get(request.url.hostname) is None:
                socket_configs = await request.url.lookup()

                for ip_addr, configs in socket_configs.items():
                        for config in configs:

                            connection = HTTP2Connection(
                                0, 
                                self.timeouts, 
                                1,
                                self.pool.reset_connections,
                                self.pool.pool_type
                            )

                            try:
                                
                                await connection.connect(
                                    request.url.hostname,
                                    ip_addr,
                                    request.url.port,
                                    config,
                                    ssl=self.ssl_context,
                                    timeout=self.timeouts.connect_timeout
                                )

                                request.url.socket_config = config
                                break

                            except Exception as e:
                                pass

                        if request.url.socket_config:
                            break

                if request.url.socket_config is None:
                        raise Exception('Err. - No socket found.')
                
                self._hosts[request.url.hostname] = request.url.ip_addr
                        
            else:
                request.url.ip_addr = self._hosts[request.url.hostname]

            if request.is_setup is False:
                request.setup()

            self.registered[request.name] = request

        except Exception as e:
            raise e

    def extend_pool(self, increased_capacity: int):
        self.pool.size += increased_capacity
        self.pool.create_pool()
    
        self.sem = Semaphore(self.pool.size)

    def shrink_pool(self, decrease_capacity: int):
        self.pool.size -= decrease_capacity
        self.pool.create_pool()

        self.sem = Semaphore(self.pool.size)

    async def execute_prepared_request(self, action: HTTP2Action) -> HTTP2ResponseFuture:
        response = HTTP2Result(action)
        response.wait_start = time.monotonic()
        self.active += 1
        
        async with self.sem:

            pipe = self.pool.pipes.pop()
            connection = self.pool.connections.pop()
        
            try:

                if action.hooks.listen:
                    event = asyncio.Event()
                    action.hooks.channel_events.append(event)
                    await event.wait()

                if action.hooks.before:
                    action = await action.hooks.before(action, response)
                    action.setup()

                response.start = time.monotonic()

                stream = await connection.connect(
                    action.url.hostname,
                    action.url.ip_addr,
                    action.url.port,
                    action.url.socket_config,
                    ssl=action.ssl_context
                )

                stream.encoder = action.hpack_encoder
     
                response.connect_end = time.monotonic()

                pipe.send_request_headers(action, stream)
  
                if action.encoded_data is not None:
                    await pipe.submit_request_body(action, stream)

                response.write_end = time.monotonic()

                await asyncio.wait_for(
                    pipe.receive_response(response, stream), 
                    timeout=self.timeouts.total_timeout
                )

                response.complete = time.monotonic()

                if action.hooks.after:
                    action = await action.hooks.after(action, response)
                    action.setup()

                if action.hooks.notify:
                    await asyncio.gather(*[
                        asyncio.create_task(
                            channel(response, action.hooks.listeners)
                        ) for channel in action.hooks.channels
                    ])

                    for listener in action.hooks.listeners: 
                        if len(listener.hooks.channel_events) > 0:
                            listener.setup()
                            event = listener.hooks.channel_events.pop()
                            if not event.is_set():
                                event.set()      

                self.pool.pipes.append(pipe)
                self.pool.connections.append(connection)
                
            except Exception as e:
                response.complete = time.monotonic()
                response.response_code = 400
                response.error = str(e)

                self.pool.reset()

            self.active -= 1
            if self.waiter and self.active <= self.pool.size:

                try:
                    self.waiter.set_result(None)
                    self.waiter = None

                except asyncio.InvalidStateError:
                    self.waiter = None

            return response

    async def close(self):
        if self.closed is False:
            await self.pool.close()
            self.closed = True



from enum import Enum


class ReporterTypes(Enum):
    AWSLambda='aws_lambda'
    AWSTimestream='aws_timestream'
    BigQuery='bigquery'
    BigTable='bigtable'
    Cassandra='cassandra'
    Cloudwatch='cloudwatch'
    CosmosDB='cosmosdb'
    CSV='csv'
    Datadog='datadog'
    DogStatsD='dogstatsd'
    GCS='gcs'
    Graphite='graphite'
    Honeycomb='honeycomb'
    InfluxDB='influxdb'
    JSON='json'
    Kafka='kafka'
    MongoDB='mongodb'
    MySQL='mysql'
    Netdata='netdata'
    NewRelic='newrelic'
    Postgres='postgres'
    Prometheus='prometheus'
    Redis='redis'
    S3='s3'
    Snowflake='snowflake'
    SQLite='sqlite'
    StatsD='satsd'
    Telegraf='telegraf'
    TelegrafStatsD='telegraf_statsd'
    TimescaleDB='timescaledb'


import asyncio
from .worker_services.worker_manager import WorkerManager


class DistributedWorkerServer:

    def __init__(self, config):
        self.manager = WorkerManager(config, config.reporter_config)
        self.loop = None
        
    def register(self):
        self.loop = asyncio.get_event_loop()
        self.loop.run_until_complete(self.manager.start_server())
        self.loop.run_until_complete(self.manager.register())

    def run(self):
        self.loop.run_until_complete(self.manager.wait())

    def kill(self):
        self.loop.run_until_complete(self.manager.stop())




from .registered_service import RegisteredService


class ServiceRegistry:

    def __init__(self, expected_leaders) -> None:
        self.services = {}

        if expected_leaders is None:
            expected_leaders = 1

        self.expected_leaders = expected_leaders

    async def register_service(self, registry_request):
        service_name = registry_request.get('service_name')
        registered_service = self.services.get(
            service_name, 
            RegisteredService(self.expected_leaders)
        )
        
        if registered_service.service_registered is False:
            self.services[service_name] = registered_service
            registered_service.service_registered = True

        await registered_service.register_node(registry_request)

        return await registered_service.get_registration_status()

    async def get_service_registration_status(self, registration_status_request):
        service_name = registration_status_request.get('service_name')
        registered_service = self.services.get(
            service_name, 
            RegisteredService(self.expected_leaders)
        )

        return await registered_service.get_registration_status()




import time
import asyncio
import uuid
from typing import Awaitable, Dict, Union, Tuple, Set
from hedra.core.engines.types.common.timeouts import Timeouts
from hedra.core.engines.types.common.ssl import get_default_ssl_context
from .connection import WebsocketConnection
from .pool import Pool
from .action import WebsocketAction
from .result import WebsocketResult
from .utils import get_header_bits, get_message_buffer_size


WebsocketResponseFuture = Awaitable[Union[WebsocketAction, Exception]]
WebsocketBatchResponseFuture = Awaitable[Tuple[Set[WebsocketResponseFuture], Set[WebsocketResponseFuture]]]


class MercuryWebsocketClient:

    __slots__ = (
        'session_id',
        'timeouts',
        'registered',
        '_hosts',
        'closed',
        'sem',
        'pool',
        'active',
        'waiter',
        'ssl_context'
    )


    def __init__(self, concurrency: int = 10 ** 3, timeouts: Timeouts = Timeouts(), reset_connections: bool=False) -> None:
        
        self.session_id = str(uuid.uuid4())
        self.timeouts = timeouts

        self.registered: Dict[str, WebsocketAction] = {}
        self._hosts = {}
        self.closed = False

        self.sem = asyncio.Semaphore(value=concurrency)
        self.pool = Pool(concurrency, reset_connections=reset_connections)
        self.pool.create_pool()
        self.active = 0
        self.waiter = None

        self.ssl_context = get_default_ssl_context()
        

        
    async def prepare(self, action: WebsocketAction) -> Awaitable[Union[WebsocketAction, Exception]]:
        try:
            if action.url.is_ssl:
                action.ssl_context = self.ssl_context

            if self._hosts.get(action.url.hostname) is None:

                    socket_configs = await action.url.lookup()
                    for ip_addr, configs in socket_configs.items():
                        for config in configs:
                            try:
                                connection = WebsocketConnection()
                                await connection.make_connection(
                                    action.url.hostname,
                                    ip_addr,
                                    action.url.port,
                                    config,
                                    ssl=action.ssl_context
                                )

                                action.url.socket_config = config
                                action.url.ip_addr = ip_addr
                                action.url.has_ip_addr = True
                                break

                            except Exception as e:
                                pass

                        if action.url.socket_config:
                            break

                    if action.url.socket_config is None:
                        raise Exception('Err. - No socket found.')
                    
                    self._hosts[action.url.hostname] = {
                        'ip_addr': action.url.ip_addr,
                        'socket_config': action.url.socket_config
                    }

            else:
                host_config = self._hosts[action.url.hostname]
                action.url.ip_addr = host_config.get('ip_addr')
                action.url.socket_config = host_config.get('socket_config')

            if action.is_setup is False:
                action.setup()

            self.registered[action.name] = action

            return action
        
        except Exception as e:
            raise e

    def extend_pool(self, increased_capacity: int):
        self.pool.size += increased_capacity
        for _ in range(increased_capacity):
            self.pool.connections.append(
                WebsocketConnection(self.pool.reset_connections)
            )
        
        self.sem = asyncio.Semaphore(self.pool.size)

    def shrink_pool(self, decrease_capacity: int):
        self.pool.size -= decrease_capacity
        self.pool.connections = self.pool.connections[:self.pool.size]
        self.sem = asyncio.Semaphore(self.pool.size)


    async def execute_prepared_request(self, action: WebsocketAction) -> WebsocketResponseFuture:

        response = WebsocketResult(action)
        response.wait_start = time.monotonic()
        self.active += 1

        async with self.sem:

            connection = self.pool.connections.pop()

            try:

                if action.hooks.listen:
                    event = asyncio.Event()
                    action.hooks.channel_events.append(event)
                    await event.wait()

                if action.hooks.before:
                    action = await action.hooks.before(action, response)
                    action.setup()

                response.start = time.monotonic()

                await connection.make_connection(
                    action.name,
                    action.url.ip_addr,
                    action.url.port,
                    ssl=action.ssl_context
                )

                response.connect_end = time.monotonic()

                connection.write(action.encoded_headers)
                
                if action.encoded_data is not None:
                    if action.is_stream:
                        action.write_chunks(connection)

                    else:
                        connection.write(action.encoded_data)

                response.write_end = time.monotonic()

                line = await asyncio.wait_for(connection.readuntil(), self.timeouts.socket_read_timeout)

                response.response_code = line
                raw_headers = b''
                async for key, value, header_line in connection.iter_headers(connection):
                    response.headers[key] = value
                    raw_headers += header_line

                if action.encoded_data is not None:
                    header_bits = get_header_bits(raw_headers)
                    header_content_length = get_message_buffer_size(header_bits)
                    
                if action.method == 'GET':
                    response.body = await asyncio.wait_for(connection.readexactly(min(16384, header_content_length)), self.timeouts.total_timeout)
                
                response.complete = time.monotonic()

                if action.hooks.after:
                    action = await action.hooks.after(action, response)
                    action.setup()

                if action.hooks.notify:
                    await asyncio.gather(*[
                        asyncio.create_task(
                            channel(response, action.hooks.listeners)
                        ) for channel in action.hooks.channels
                    ])

                    for listener in action.hooks.listeners: 
                        if len(listener.hooks.channel_events) > 0:
                            listener.setup()
                            event = listener.hooks.channel_events.pop()
                            if not event.is_set():
                                event.set()   

                self.pool.connections.append(connection)

            except Exception as e:
                response.complete = time.monotonic()
                response.error = str(e)

                self.pool.connections.append(
                    WebsocketConnection(reset_connection=self.pool.reset_connections) 
                )

            self.active -= 1
            if self.waiter and self.active <= self.pool.size:

                try:
                    self.waiter.set_result(None)
                    self.waiter = None

                except asyncio.InvalidStateError:
                    self.waiter = None 

            return response

    async def close(self):
        if self.closed is False:
            await self.pool.close()
            self.closed = True




import uvloop
import traceback
uvloop.install()

import os
import click
import asyncio
from typing import TypeVar
from hedra.core.engines.types.registry import registered_engines
from hedra.core.engines.types.common.types import RequestTypes
from hedra.core.engines.types.common.timeouts import Timeouts
from hedra.core.engines.types.grpc import GRPCAction, MercuryGRPCClient
from hedra.core.engines.types.graphql import GraphQLAction, MercuryGraphQLClient
from hedra.core.engines.types.graphql_http2 import GraphQLHTTP2Action, MercuryGraphQLHTTP2Client
from hedra.core.engines.types.http import HTTPAction, MercuryHTTPClient
from hedra.core.engines.types.http2 import HTTP2Action, MercuryHTTP2Client
from hedra.core.engines.types.playwright import MercuryPlaywrightClient
from hedra.core.engines.types.udp import UDPAction, MercuryUDPClient
from hedra.core.engines.types.websocket import WebsocketAction, MercuryWebsocketClient
from hedra.logging import (
    HedraLogger,
    LoggerTypes,
    logging_manager
)


T = TypeVar(
    'T', 
    MercuryGRPCClient,
    MercuryGraphQLHTTP2Client, 
    MercuryGraphQLClient, 
    MercuryHTTP2Client, 
    MercuryHTTPClient,
    MercuryPlaywrightClient,
    MercuryUDPClient,
    MercuryWebsocketClient
)

@click.command(help="Ping the specified uri to ensure it can be reached.")
@click.argument('uri')
@click.option('--engine', default='http', type=str)
@click.option('--timeout', default=60, type=int)
@click.option('--log-level', default='info', type=str)
def ping(uri: str, engine: str, timeout: int, log_level: str):

    logging_manager.disable(
        LoggerTypes.HEDRA, 
        LoggerTypes.DISTRIBUTED,
        LoggerTypes.FILESYSTEM,
        LoggerTypes.DISTRIBUTED_FILESYSTEM
    )
    logging_manager.update_log_level(log_level)

    logger = HedraLogger()
    logger.initialize()
    logging_manager.logfiles_directory = os.getcwd()

    engine_types_map = {
        'http': RequestTypes.HTTP,
        'http2': RequestTypes.HTTP2,
        'grpc': RequestTypes.GRPC,
        'graphql': RequestTypes.GRAPHQL,
        'graphql-http2': RequestTypes.GRAPHQL_HTTP2,
        'playwright': RequestTypes.PLAYWRIGHT,
        'websocket': RequestTypes.WEBSOCKET
    }

    logger['console'].sync.info(f'Pinging target - {uri} - using engine - {engine}.')
    logger['console'].sync.debug(f'Pinging with timeout of - {timeout} - seconds.')

    engine_type = engine_types_map.get(engine, RequestTypes.HTTP)

    asyncio.run(ping_target(uri, engine_type, timeout, logger))
    

async def ping_target(uri: str, engine_type: RequestTypes, timeout: int, logger: HedraLogger):

    action_name = f'ping_{uri}'

    timeouts = Timeouts(
        connect_timeout=timeout,
        total_timeout=timeout
    )
    
    selected_engine: T  = registered_engines.get(engine_type, MercuryHTTPClient)(
        concurrency=1,
        timeouts=timeouts,
        reset_connections=False
    )

    action_types = {
        RequestTypes.HTTP: HTTPAction(
            action_name,
            uri
        ),
        RequestTypes.HTTP2: HTTP2Action(
            action_name,
            uri
        ),
        RequestTypes.GRAPHQL: GraphQLAction(
            action_name,
            uri
        ),
        RequestTypes.GRAPHQL_HTTP2: GraphQLHTTP2Action(
            action_name,
            uri
        ),
        RequestTypes.GRPC: GRPCAction(
            action_name,
            uri
        ),
        RequestTypes.PLAYWRIGHT: HTTPAction(
            action_name,
            uri
        ),
        RequestTypes.UDP: UDPAction(
            action_name,
            uri
        ),
        RequestTypes.WEBSOCKET: WebsocketAction(
            action_name,
            uri
        )
    }

    action = action_types.get(
        engine_type, 
        HTTPAction(
            action_name,
            uri
        )
    )

    if engine_type == RequestTypes.PLAYWRIGHT:
        selected_engine = MercuryHTTPClient(timeouts=timeouts)

    try:
        
        await logger['console'].aio.debug(f'Preparing to connect to - {uri}.')

        action.setup()
        await selected_engine.prepare(action)

        await logger['console'].aio.info(f'Successfully connected to - {uri}!\n')

    except Exception:
        ping_error = traceback.format_exc().split('\n')[-2]
        await logger['console'].aio.error(f'Error - could not ping - {uri}.\nEncountered - {str(ping_error)} - error.\n')



import uuid
from typing import Any, Dict, List, Union
from hedra.reporting.tags import Tag
from .metrics_group import MetricsGroup


class MetricsSet:

    def __init__(
        self,
        name: str,
        source: str,
        stage: str, 
        metrics_data: Dict[str, Dict[str, Any]],
        tags: Dict[str, str]
    ) -> None:

        self.metrics_set_id = str(uuid.uuid4())
        self.groups: Dict[str, MetricsGroup] = {}

        self.name = name
        self.source = source
        self.stage = stage
        self._raw_metrics = metrics_data
        self._raw_tags = tags

        self.fields = [
            'name',
            'stage'
        ]
        self.values = [
            name,
            stage
        ]

        self.common_stats = {
            'total': metrics_data.get('total'),
            'succeeded': metrics_data.get('succeeded'),
            'failed': metrics_data.get('failed'),
            'actions_per_second': metrics_data.get('actions_per_second')
        }

        self.custom_metrics = metrics_data.get('custom')
        self.errors: List[Dict[str, Union[str, int]]] = metrics_data.get('errors')
        self.tags = [
            Tag(tag_name, tag_value) for tag_name, tag_value in tags.items()
        ]

        metrics_groups = metrics_data.get('groups', {})
        for group_name, group in metrics_groups.items():
            self.groups[group_name] = MetricsGroup(
                name,
                source,
                stage,
                group_name,
                group,
                self.common_stats
            )

        record_fields = self.groups.get('total').fields
        custom_fields = list(self.groups.get(
            'total'
        ).custom_fields.keys())

        self.fields.extend(record_fields)
        self.fields.extend(custom_fields)

        self.quantiles = list(self.groups.get(
            'total'
        ).quantiles.keys())

        self.custom_field_names = self.custom_fields = self.groups.get(
            'total'
        ).custom_field_names


        self.custom_fields = self.groups.get(
            'total'
        ).custom_fields

        self.custom_schemas = self.groups.get(
            'total'
        ).custom_schemas

        self.stats_fields = list(self.groups.get(
            'total'
        ).stats.keys())

        

    def serialize(self):
        return {
            'name': self.name,
            'source': self.source,
            'stage': self.stage,
            'metrics_data': self._raw_metrics,
            'tags': self._raw_tags
        }



from typing import Generic, TypeVar
from hedra.core.engines.types.common.base_result import BaseResult
from .action import Action

R = TypeVar('R')

class Result(BaseResult, Generic[R]):

    __slots__ = (
        'times',
        "type"
    )

    def __init__(self, action: Action, error: Exception = None) -> None:
        super().__init__(
            action.name, 
            action.source,
            action.metadata.user,
            action.metadata.tags,
            action.plugin_type,
            action.hooks.checks,
            error
        )

        self.times = {}

    def as_timings(self):
        return {
            'total': self.times['complete'] - self.times['start']
        }


import uuid
from typing import Coroutine
from hedra.core.graphs.stages.base.stage import Stage


class Transition:

    def __init__(self, transition: Coroutine, from_stage: Stage, to_stage: Stage) -> None:
        self.transition_id = str(uuid.uuid4())
        self.transition = transition
        self.from_stage = from_stage
        self.to_stage = to_stage

    def assemble(self):
        return self.transition(self.from_stage, self.to_stage)


from hedra.command_line import CommandLine
from hedra.core.personas.parallel import calculate_total_actions


def generate_configs(pool_size, cli):
    configs = []

    total_actions = calculate_total_actions(
            total_actions=cli.executor_config.get('total_actions'),
            total_time=cli.executor_config.get('total_time'),
            batch_size=cli.executor_config.get('batch_size')
        )

    pool_bin_size = int(total_actions/pool_size)
    last_bin_size = pool_bin_size

    if total_actions % pool_size > 0:
        last_bin_size = total_actions % pool_size

    for _ in range(pool_size):
        command_line = CommandLine()
        command_line = command_line.copy(cli)
        command_line.executor_config['total_actions'] = pool_bin_size
        
        configs.append(command_line)

    configs[len(configs)-1].executor_config['total_actions'] = last_bin_size

    return configs



import os
from hedra.core.graphs.stages import (
    Submit
)

from hedra.reporting.types import (
    GoogleCloudStorageConfig
)


class SubmitGoogleCloudStorageResultsStage(Submit):
    config=GoogleCloudStorageConfig(
        service_account_json_path=os.getenv('GOOGLE_CLOUD_ACCOUNT_JSON_PATH', ''),
        bucket_namespace='results',
        events_bucket='events',
        metrics_bucket='metrics'
    )


import asyncio
import psutil
import time
import uuid
import statistics
from typing import List, Dict, Union, Any
from asyncio import Task
from hedra.core.personas.types.default_persona import DefaultPersona
from hedra.core.engines.client.config import Config
from hedra.core.personas.types.types import PersonaTypes



async def cancel_pending(pend: Task):
    try:
        pend.cancel()
        if not pend.cancelled():
            await pend

        return pend
    
    except asyncio.CancelledError as cancelled_error:
        return cancelled_error


class CyclicNoWaitPersona(DefaultPersona):

    def __init__(self, config: Config):
        super().__init__(config)

        self.persona_id = str(uuid.uuid4())
        self.type = PersonaTypes.NO_WAIT

    async def generator(self, total_time):
        elapsed = 0
        idx = 0

        start = time.monotonic()
        while elapsed < total_time:
            yield idx%self.actions_count
            
            await asyncio.sleep(0)
            elapsed = time.monotonic() - start
            idx += 1



from hedra.core.graphs.hooks.registry.registry_types.hook import Hook
from hedra.core.graphs.hooks.hook_types.hook_type import HookType


class HookSetupError(Exception):

    def __init__(self, hook: Hook, hook_type: HookType, message: str) -> None:

        hook_type = hook_type.name.lower()
        super().__init__(
            f'Hook Error - @{hook_type} hook {hook.shortname} from stage {hook.stage}\nEncountered exception - {message} - while attempting to setup hook.'
        )




from ensurepip import version
import json
from gzip import decompress as gzip_decompress
from typing import List, Union
from zlib import decompress as zlib_decompress
from hedra.core.engines.types.common.types import RequestTypes
from hedra.core.engines.types.common.base_result import BaseResult
from .action import HTTPAction

class HTTPResult(BaseResult):

    __slots__ = (
        'action_id',
        'url',
        'ip_addr',
        'method',
        'path',
        'params',
        'query',
        'hostname',
        'headers',
        'body',
        'response_code',
        '_version',
        '_reason',
        '_status'
    )

    def __init__(self, action: HTTPAction, error: Exception=None) -> None:

        super(
            HTTPResult,
            self
        ).__init__(
            action.action_id,
            action.name,
            action.url.hostname,
            action.metadata.user,
            action.metadata.tags,
            RequestTypes.HTTP,
            action.hooks.checks,
            error
        )

        self.url = action.url.full
        self.ip_addr = action.url.ip_addr
        self.method = action.method
        self.path = action.url.path
        self.params = action.url.params
        self.query = action.url.query
        self.hostname = action.url.hostname

        self.headers = {}
        self.body = bytearray()
        self.response_code = None
        self._version = None
        self._reason = None
        self._status = None

    @property
    def content_type(self):
        return self.headers.get(b'content-type')

    @property
    def compression(self):
        return self.headers.get(b"content-encoding")

    @property
    def size(self):
        if self.headers.get(b'content-length'):
            return int(self.headers.get(b'content-length'))
        
        elif self.body:
            return len(self.body)
        
        else:
            return 0

    @property
    def data(self) -> Union[str, dict, None]:
        data = self.body
        try:
            if self.compression == b"gzip":
                data = gzip_decompress(self.body)
            elif self.compression == b"deflate":
                data = zlib_decompress(self.body)

            if self.content_type == b"application/json":
                data = json.loads(self.body)
            
            elif isinstance(self.body, (bytes, bytearray)):
                data = str(data.decode())

        except Exception:
            pass

        return data

    @property
    def version(self) -> Union[str, None]:
        try:
            if self._version is None and isinstance(self.response_code, (bytes, bytearray)):
                status_string: List[bytes] = self.response_code.split()
                self._version = status_string[0].decode()
        except Exception:
            pass
            
        return self._version

    @version.setter
    def version(self, new_version: str):
        self._version = new_version

    @property
    def status(self) -> Union[int, None]:
        try:

            if self._status is None and isinstance(self.response_code, (bytes, bytearray)):
                status_string: List[bytes] = self.response_code.split()
                self._status = int(status_string[1])

        except Exception:
            pass

        return self._status

    @status.setter
    def status(self, new_status: int):
        self._status = new_status

    @property
    def reason(self) -> Union[str, None]:

        try:

            if self._reason is None and isinstance(self.response_code, (bytes, bytearray)):
                status_string: List[bytes] = self.response_code.split()
                self._reason = status_string[2].decode()

        except Exception:
            pass
        
        return self._reason

    @reason.setter
    def reason(self, new_reason):
        self._reason = new_reason

    def to_dict(self):

        encoded_headers = {
            str(k.decode()): str(v.decode()) for k, v in self.headers.items()
        }

        base_result_dict = super().to_dict()

        data = self.data
        if isinstance(data, bytes) or isinstance(data, bytearray):
            data = str(data.decode())
        
        return {
            'url': self.url,
            'method': self.method,
            'path': self.path,
            'params': self.params,
            'query': self.query,
            'type': self.type,
            'headers': encoded_headers,
            'data': data,
            'tags': self.tags,
            'user': self.user,
            'error': str(self.error),
            'status': self.status,
            'reason': self.reason,
            **base_result_dict
        }


from typing import Type, Callable, Awaitable, Any
from hedra.core.graphs.hooks.hook_types.hook_type import HookType
from .hook import Hook


class ValidateHook(Hook):

    def __init__(
        self, 
        name: str, 
        shortname: str, 
        call: Callable[..., Awaitable[Any]], 
        *names: str
    ) -> None:
        super().__init__(
            name, 
            shortname, 
            call, 
            hook_type=HookType.VALIDATE
        )
        
        self.call: Type[self._call] = self._call
        self.names = list(set(names))


from typing import List, Tuple, Union, Dict
from .types import (
    SHGOptimizer,
    DualAnnealingOptimizer,
    DifferentialEvolutionOptimizer
)


registered_algorithms = {
        'shg': SHGOptimizer,
        'dual-annealing': DualAnnealingOptimizer,
        'diff-evolution': DifferentialEvolutionOptimizer
}


def get_algorithm(
    algorithm_type: str,
    config: Dict[str, Union[List[Tuple[Union[int, float]]], int]]
):
    return registered_algorithms.get(
        algorithm_type,
        SHGOptimizer
    )(config)


from typing import Dict, List, Optional


class Metadata:

    __slots__ = (
        'user',
        'tags'
    )

    def __init__(self, user: Optional[str] = None, tags: List[Dict[str, str]] = []) -> None:
        self.user = user
        self.tags = tags

    def __aiter__(self):
        for tag in self.tags:
            yield tag

    def update_tags(self, user: str=None, tags: List[Dict[str, str]]=[]):
        if user:
            self.user = user
            
        self.tags.extend(tags)

    def tags_to_string_list(self):
        return [
           f'{tag.get("name")}:{tag.get("value")}'  for tag in self.tags
        ]


from .plugins import (
    CustomEngine,
    CustomOptimizer,
    CustomPersona,
    CustomReporter
)

from hedra.plugins.types.common.registrar import plugin_registrar
from hedra.projects.generation.generator import Generator


class PluginGenerator(Generator):

    def __init__(self) -> None:
        super().__init__({
            'engine': CustomEngine,
            'optimizer': CustomOptimizer,
            'persona': CustomPersona,
            'reporter': CustomReporter
        }, plugin_registrar.module_paths)

    def generate_plugin(self, plugin_type: str) -> str:
        modules = self.gather_required_items(plugin_type)
        self.collect_imports(plugin_type, modules)

        self.serialize_items()

        serialized_imports = '\n'.join([
            *self.serialized_global_imports,
            *self.serialized_local_imports,

        ])

        return '\n\n'.join([
            serialized_imports,
            *self.serialized_locals
        ])


from typing import Dict, List, Any

class ContextConfig:

    __slots__ = (
        'data',
        'options'
    )

    def __init__(
        self, 
        browser_type: str='chromium', 
        device_type: str=None, 
        locale: str=None, 
        geolocation: Dict[str, float]=None, 
        permissions: List[str]=[], 
        color_scheme: str=None,
        options: Dict[str, Any]={}
    ) -> None:
        self.data = {
            'browser_type': browser_type,
            'device_type': device_type,
            'locale': locale,
            'geolocation': geolocation,
            'permissions': permissions,
            'color_scheme': color_scheme
        }

        self.options = options


import json
from typing import Dict, Any, Union
from hedra.core.engines import registered_engines
from hedra.core.graphs.hooks.registry.registrar import registrar
from hedra.core.engines.client.config import Config
from hedra.core.engines.types.common.timeouts import Timeouts
from hedra.core.graphs.hooks.registry.registry_types import ActionHook, TaskHook
from hedra.core.personas.types.default_persona import DefaultPersona
from hedra.plugins.types.engine.engine_plugin import EnginePlugin
from hedra.plugins.types.engine.action import Action
from hedra.logging import HedraLogger


async def assemble_plugin_action(
    hook: Union[ActionHook, TaskHook],
    hook_action: Dict[str, Any],
    persona: DefaultPersona,
    config: Config,
    metadata_string: str
):

    logger = HedraLogger()
    logger.initialize()

    action_name = hook_action.get('name')
    action_fields: Dict[str, Any] = hook_action.get('fields', {})

    action_hooks = hook_action.get('hooks', {})
    action_hook_names = action_hooks.get('names')

    before_hook_name = action_hook_names.get('before')
    after_hook_name = action_hook_names.get('after')
    check_hook_names = action_hook_names.get('checks', [])
    listener_names = action_hook_names.get('listeners', [])
    channel_hook_names = action_hook_names.get('channels', [])


    plugin_type: str = hook_action.get('plugin_type')
    plugin_type_name = plugin_type.capitalize()

    await logger.filesystem.aio['hedra.core'].debug(
        f'{metadata_string} - Assembling {plugin_type_name} Hook - {action_name}'
    )

    if before_hook_name:
        await logger.filesystem.aio['hedra.core'].debug(
            f'{metadata_string} - {plugin_type_name} Hook - {action_name} - Found Before Hook - {before_hook_name}'
        )

    if after_hook_name:
        await logger.filesystem.aio['hedra.core'].debug(
            f'{metadata_string} - {plugin_type_name} Hook - {action_name} - Found After Hook - {after_hook_name}'
        )

    for check_hook_name in check_hook_names:
        await logger.filesystem.aio['hedra.core'].debug(
            f'{metadata_string} - {plugin_type_name} Hook - {action_name} - Found Check Hook - {check_hook_name}'
        )

    for channel_hook_name in channel_hook_names:
        await logger.filesystem.aio['hedra.core'].debug(
            f'{metadata_string} - {plugin_type_name} Hook - {action_name} - Found Channel Hook - {channel_hook_name}'
        )

    for listener_name in listener_names:
        await logger.filesystem.aio['hedra.core'].debug(
            f'{metadata_string} - {plugin_type_name} Hook - {action_name} - Found Listener - {listener_name}'
        )

    timeouts: Timeouts = hook_action.get('timeouts', {})
    reset_connections = hook_action.get('reset_connections')

    await logger.filesystem.aio['hedra.core'].debug(f'{metadata_string} - {plugin_type_name} Hook - {action_name} - Connect Timeout: {timeouts.connect_timeout}')
    await logger.filesystem.aio['hedra.core'].debug(f'{metadata_string} - {plugin_type_name} Hook - {action_name} - Read Timeout: {timeouts.socket_read_timeout}')
    await logger.filesystem.aio['hedra.core'].debug(f'{metadata_string} - {plugin_type_name} Hook - {action_name} - Total Timeout: {timeouts.total_timeout}')
    await logger.filesystem.aio['hedra.core'].debug(f'{metadata_string} - {plugin_type_name} Hook - {action_name} - Reset Connections: {reset_connections}')

    plugin_config = Config(
        batch_size=persona.batch.size,
        connect_timeout=timeouts.connect_timeout,
        request_timeout=timeouts.total_timeout,
        reset_connections=reset_connections
    )

    plugin: EnginePlugin = registered_engines.get(plugin_type)(plugin_config)
    hook.session = plugin

    hook.action: Action = plugin.action(**{
        'name': action_name,
        **action_fields,
        **hook_action.get('metadata')
    })

    for field_name, field_value in action_fields.items():
        serialized_field = json.dumps({field_name: field_value})
        await logger.filesystem.aio['hedra.core'].debug(f'{metadata_string} - {plugin_type_name} Hook - {action_name} - {serialized_field}')

    hook.action.use_security_context = hook_action.get('use_security_context', False)
    hook.action.security_context = hook_action.get('security_context', False)
    hook.action.plugin_type = plugin_type

    await logger.filesystem.aio['hedra.core'].debug(f'{metadata_string} - {plugin_type_name} Hook - {action_name} - Use Security Context: {hook.action.use_security_context}')
    await logger.filesystem.aio['hedra.core'].debug(f'{metadata_string} - {plugin_type_name} Hook - {action_name} - User: {hook.action.metadata.user}')

    for tag in hook.action.metadata.tags_to_string_list():
        await logger.filesystem.aio['hedra.core'].debug(f'{metadata_string} - {plugin_type_name} Hook - {action_name} - Tag: {tag}')

    hook.action.hooks.notify = action_hooks.get('notify', False)
    hook.action.hooks.listen = action_hooks.get('listen', False)

    if before_hook_name:
        before_hook = registrar.all.get(before_hook_name)
        hook.action.hooks.before = before_hook.call

    if after_hook_name:
        after_hook = registrar.all.get(after_hook_name)
        hook.action.hooks.after = after_hook.call

    hook.action.hooks.checks = []
    hook.action.hooks.checks.extend(check_hook_names)

    hook.action.hooks.channels = []
    for channel_hook_name in channel_hook_names:
        channel = registrar.all.get(channel_hook_name)
        hook.action.hooks.channels.append(channel.call) 

    hook.action.hooks.listeners = listener_names

    await logger.filesystem.aio['hedra.core'].debug(f'{metadata_string} - {plugin_type_name} Hook - {action_name} - Setup complete')

    return hook


import json
from typing import Generic, TypeVar
from hedra.reporting.events.types.base_event import BaseEvent


T = TypeVar('T')

class Event(BaseEvent, Generic[T]):

    __slots__ = (
        'fields',
        'type'
    )

    def __init__(self, result: T) -> None:
        super().__init__(result)

        self.timings = result.as_timings()
        self.time = result.times['complete'] - result.times['start']

    def to_dict(self):
        return {
            'name': self.name,
            'stage': self.stage,
            'shortname': self.shortname,
            'checks': [check.__name__ for check in self.checks],
            'error': str(self.error),
            'time': self.time,
            'type': self.type,
            'source': self.source,
        }

    def serialize(self):
        return json.dumps(self.to_dict())